from ..base import BaseTransformer
from datetime import datetime
import holidays
from neurostats_API.utils import StatsProcessor
import numpy as np
import pandas as pd
import pytz


class BaseFinanceOverviewTransformer(BaseTransformer):

    def __init__(self, ticker, company_name, zone):
        super().__init__(ticker, company_name, zone)
        self.tw_company_list = StatsProcessor.load_json(
            "company_list/ticker_index_industry_map.json"
        )
        self.us_company_list = StatsProcessor.load_json(
            "company_list/us_TradingView_list.json"
        )

    def _get_latest_date(self, tech_data):
        latest_date = tech_data.iloc[-1]['date']
        if (latest_date is None):
            str_date = datetime.today().strftime("%Y-%m-%d")
        if (isinstance(latest_date, datetime)):
            str_date = latest_date.strftime("%Y-%m-%d")
        elif (isinstance(latest_date, str)):
            str_date = latest_date

        return f"{str_date} [GMT+8]"

    def _get_latest_close(self, tech_data):
        latest_close = tech_data.iloc[-1]['close']
        latest_close = str(round(latest_close, 2))
        return latest_close

    def _get_latest_offset(self, tech_data):
        latest_close = tech_data.iloc[-1]['close']
        previous = tech_data.iloc[-2]['close']

        offset = latest_close - previous
        percentage = offset / previous

        offset = str(np.round(offset, 2))

        return {'value': offset, "percentage": f"{percentage:.2f}%"}

    def _get_category(self):

        def tw_category(ticker):
            return self.tw_company_list.get(ticker, {}).get('industry')

        def us_category(ticker):
            return self.us_company_list.get(ticker, {}).get('en_industry')

        fn_map = {'tw': tw_category, 'us': us_category}

        get_fn = fn_map.get(self.zone)
        return get_fn(self.ticker)

    def _get_latest_volume(self, tech_data):
        try:
            volume = round(tech_data[-1]['volume'], 2)
            return str(volume)
        except Exception as e:
            return "Error"

    def _get_TTM_data(self, datas):
        dfs = pd.DataFrame(datas)
        TTM_data = dfs.mean()

        return TTM_data

    def _process_TTM_tw(self, datas):
        return datas

    def _get_latest_average_volume(self, tech_data, length=30):
        try:
            tech_data = pd.DataFrame(tech_data).tail(length)

            value = tech_data['volume'].mean().item()
            value = round(value, 2)

            return str(value)

        except Exception as e:
            return "Error"

    def _get_market_capitalization(self, daily_data, seasonal_data):
        try:
            latest_close = float(self._get_latest_close(daily_data))
            latest_common_share = seasonal_data[-1]['Common Stock']

            percentage = round((latest_close * latest_common_share), 2)

            return str(percentage)

        except Exception as e:
            return "Error"

    def _get_PE(self, daily_data, TTM_data):
        try:
            latest_close = float(self._get_latest_close(daily_data))
            TTM_EPS = TTM_data['Diluted EPS']

            value = round((latest_close / TTM_EPS), 2)

            return str(value)

        except Exception as e:
            return "Error"

    def _get_PS(self, daily_data, TTM_data, seasonal_data):
        try:
            market_capitalzation = self._get_market_capitalization(
                daily_data, seasonal_data
            )
            market_capitalzation = float(market_capitalzation)
            TTM_revenue = TTM_data['Total Revenue']

            value = round((market_capitalzation / TTM_revenue), 2)

            return str(value)
        except Exception as e:
            return "Error"

    def _is_us_market_open(self):

        taiwan_dt = datetime.now(pytz.timezone('Asia/Taipei'))

        # 轉成美東時間（會自動處理夏令時間）
        eastern = pytz.timezone('US/Eastern')
        us_dt = taiwan_dt.astimezone(eastern)

        # 假日判斷
        us_holidays = holidays.NYSE(years=us_dt.year)
        if us_dt.date() in us_holidays:
            return False

        # 週末
        if us_dt.weekday() >= 5:    # 5 = Saturday, 6 = Sunday
            return False

        # 判斷是否在開盤時間
        market_open = us_dt.replace(hour=9, minute=30, second=0, microsecond=0)
        market_close = us_dt.replace(hour=16, minute=0, second=0, microsecond=0)

        return market_open <= us_dt <= market_close

    def _process_us_format(self, fetched_data):
        """
        主要用於report generator
        """
        table_dict = {}
        for data in fetched_data:
            year, season, balance_sheet = data['year'], data['season'], data[
                'balance_sheet']
            time_index = f"{year}Q{season}"

            table_dict[time_index] = balance_sheet

        stats_df = pd.DataFrame.from_dict(table_dict)
        return stats_df.T


class FinanceOverviewProcessor(StatsProcessor):

    @classmethod
    def process_rate(cls, finance_dict):
        for key in finance_dict:
            if ('YoY' in key):
                finance_dict[key] = StatsProcessor.cal_percentage(
                    finance_dict[key]
                )
            elif ("rate" in key or 'ratio' in key):
                finance_dict[key] = StatsProcessor.cal_non_percentage(
                    finance_dict[key], to_str=True, postfix='%'
                )
            else:
                finance_dict[key] = StatsProcessor.cal_non_percentage(
                    finance_dict[key]
                )

    @classmethod
    def process_thousand_dollar(cls, finance_dict):
        process_index = [
            "revenue", "gross_profit", "operating_income", "net_income",
            "operating_cash_flow", "invest_cash_flow", "financing_cash_flow",
            "fcf", 'current_assets', 'current_liabilities',
            'non_current_assets', 'non_current_liabilities', 'total_assets',
            "total_liabilities", "equity"
        ]

        for index in process_index:
            try:
                finance_dict[index] = StatsProcessor.cal_non_percentage(
                    finance_dict[index], postfix="千元"
                )
            except Exception as e:
                finance_dict[index] = None

    @classmethod
    def process_all(cls, finance_dict):
        methods = [
            cls.cal_EBIT,
            cls.cal_share_outstanding,
            cls.cal_fcf,
            cls.cal_interest_bearing_debt,
            cls.cal_revenue_per_share,
            cls.cal_gross_per_share,
            cls.cal_operating_income_per_share,
            cls.cal_operating_cash_flow_per_share,
            cls.fcf_per_share,
            cls.cal_roa,
            cls.cal_roe,
            cls.cal_gross_over_asset,
            cls.cal_roce,
            cls.cal_gross_profit_marginal,
            cls.cal_operation_profit_rate,
            cls.cal_operating_cash_flow_profit_rate,
            cls.cal_dso,
            cls.cal_account_receive_over_revenue,
            cls.cal_dpo,
            cls.cal_inventories_cycle_ratio,
            cls.cal_dio,
            cls.cal_inventories_revenue_ratio,
            cls.cal_cash_of_conversion_cycle,
            cls.cal_asset_turnover,
            cls.cal_application_turnover,
            cls.cal_current_ratio,
            cls.cal_quick_ratio,
            cls.cal_debt_to_equity_ratio,
            cls.cal_net_debt_to_equity_ratio,
            cls.cal_interest_coverage_ratio,
            cls.cal_debt_to_operating_cash_flow,
            cls.cal_debt_to_free_cash_flow,
            cls.cal_cash_flow_ratio,
        ]

        for method in methods:
            method(finance_dict)

    @classmethod
    def cal_EBIT(cls, finance_dict):
        """
        計算EBIT
        EBIT = 營業收入 - 營業成本 - 營業費用 - 所得稅費用
        """
        try:
            EBIT = (
                finance_dict['revenue'] - finance_dict['operating_cost'] -
                finance_dict['operating_expenses'] - finance_dict['tax_fee']
            )
            finance_dict['EBIT'] = StatsProcessor.cal_non_percentage(EBIT)
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['EBIT'] = None
            # print(f"Error calculating EBIT: {e}")

    @classmethod
    def cal_fcf(cls, finance_dict):
        """
        計算自由現金流(FCF):
        自由現金流 =  營業現金流 + 投資現金流
        """
        try:
            fcf = (
                finance_dict["operating_cash_flow"] +
                finance_dict["financing_cash_flow"]
            )
            finance_dict["fcf"] = StatsProcessor.cal_non_percentage(fcf)
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['fcf'] = None
            # print(f"Error calculating FCF: {e}")

    @classmethod
    def cal_interest_bearing_debt(cls, finance_dict):
        """
        計算有息負債
        短期借款+長期借款
        """
        finance_dict['interest_bearing_debt'] = 0.0

        try:
            finance_dict['interest_bearing_debt'] += finance_dict[
                'short_term_liabilities']
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['interest_bearing_debt'] += 0.0
        try:
            finance_dict['interest_bearing_debt'] += finance_dict[
                'long_term_liabilities']
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['interest_bearing_debt'] += 0.0

    @classmethod
    def cal_share_outstanding(cls, finance_dict):
        """
        計算流通股數
        流通股數 = 本期淨利 ÷ 基本每股盈餘
        """
        try:
            finance_dict["share_outstanding"] = (
                finance_dict['net_income'] / finance_dict['eps']
            )
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['share_outstanding'] = None
            # print(f"share_outstanding failed because of {str(e)}")

    @classmethod
    def cal_revenue_per_share(cls, finance_dict):
        """
        計算每股營收
        每股營收 = 營業收入 / 在外流通股數
        """
        try:
            revenue_per_share = (
                finance_dict['revenue'] / finance_dict['share_outstanding']
            )
            finance_dict['revenue_per_share'
                         ] = StatsProcessor.cal_non_percentage(
                             revenue_per_share, False
                         )
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['revenue_per_share'] = None
            # print(f"revenue_per_share failed because of {str(e)}")

    @classmethod
    def cal_gross_per_share(cls, finance_dict):
        """
        計算每股毛利
        = （當期營業毛利）÷（當期在外流通股數）
        """
        if ('gross_profit' not in finance_dict.keys()):
            try:
                finance_dict['gross_profit'] = (
                    finance_dict['revenue'] - finance_dict['operating_cost']
                )
            except:
                finance_dict['gross_profit'] = None
        try:
            gross_per_share = (
                finance_dict['gross_profit'] / finance_dict['share_outstanding']
            )
            finance_dict['gross_per_share'] = StatsProcessor.cal_non_percentage(
                gross_per_share, False
            )

        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['gross_per_share'] = None
            # print(f"gross_per_share failed because of {str(e)}")

    @classmethod
    def cal_operating_income_per_share(cls, finance_dict):
        """
        計算每股營業利益
        每股營業利益= （當期營業利益）÷（當期在外流通股數）
        """
        try:
            operating_income_per_share = (
                finance_dict['operating_income'] /
                finance_dict['share_outstanding']
            )
            finance_dict['operating_income_per_share'
                         ] = StatsProcessor.cal_non_percentage(
                             operating_income_per_share
                         )
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['operating_income_per_share'] = None
            # print(f"operating_income_per_share failed because of {str(e)}")

    @classmethod
    def cal_operating_cash_flow_per_share(cls, finance_dict):
        """
        計算每股營業現金流
        = (當期營業現金流) ÷（當期在外流通股數）
        """
        try:
            operating_cash_flow_per_share = (
                finance_dict["operating_cash_flow"] /
                finance_dict['share_outstanding']
            )
            finance_dict["operating_cash_flow_per_share"
                         ] = StatsProcessor.cal_non_percentage(
                             operating_cash_flow_per_share
                         )
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['operating_cash_flow_per_share'] = None
            # print(f'operating_cash_flow_per_share because of {str(e)}')

    @classmethod
    def fcf_per_share(cls, finance_dict):
        """
        計算每股自由現金流
        每股自由現金流 = (當期自由現金流) ÷（當期在外流通股數）
        """
        try:
            fcf_per_share = (
                finance_dict['fcf'] / finance_dict['share_outstanding']
            )
            finance_dict['fcf_per_share'] = StatsProcessor.cal_non_percentage(
                fcf_per_share
            )
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['fcf_per_share'] = None
            # print(f"fcf_per_share failed because of {str(e)}")

# 盈利能力

    @classmethod
    def cal_roa(cls, finance_dict):
        """
        計算資產報酬率(ROA)
        ROA = [ 本期淨利 + 利息費用 × (1-有效稅率) ] ÷（資產總額）
        """
        try:
            roa = (
                finance_dict['net_income'] + finance_dict['interest'] +
                (1 * 0.1)    # 有效稅率需要改，這裡先設0.1
            ) / finance_dict['inventories']

            finance_dict["roa"] = StatsProcessor.cal_percentage(roa)
        except Exception as e:
            finance_dict["roa"] = None

    @classmethod
    def cal_roe(cls, finance_dict):
        """
        計算股東權益報酬率(ROE)
        ROE = (本期淨利) ÷（權益總額）
        """
        try:
            roe = (finance_dict['net_income'] / finance_dict['equity'])
            finance_dict['roe'] = StatsProcessor.cal_percentage(roe)
        except Exception as e:
            finance_dict['roe'] = None

    @classmethod
    def cal_gross_over_asset(cls, finance_dict):
        """
        計算營業毛利/總資產
        """
        try:
            gross_over_asset = (
                finance_dict['gross_profit'] / finance_dict['total_assets']
            )
            finance_dict['gross_over_asset'] = StatsProcessor.cal_percentage(
                gross_over_asset
            )
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['gross_over_asset'] = None
            # print(f"營業毛利/總資產 failed because of {str(e)}")

    @classmethod
    def cal_roce(cls, finance_dict):
        """
        計算資本運用報酬率(ROCE):
        ROCE = (稅前淨利＋利息費用) / (資產總額－流動負債)
        """
        try:
            roce = (
                (
                    finance_dict['net_income_before_tax'] +
                    finance_dict['interest']
                ) / (
                    finance_dict['total_assets'] -
                    finance_dict['current_liabilities']
                )
            )
            finance_dict['roce'] = StatsProcessor.cal_percentage(roce)

        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['roce'] = None
            # print(f"ROCE failed because of {str(e)}")

    @classmethod
    def cal_gross_profit_marginal(cls, finance_dict):
        """
        計算營業毛利率(gross profit margin)
        營業毛利率 = 營業毛利 ÷ 營業收入
        """
        try:
            gross_profit_margin = (
                finance_dict['gross_profit'] / finance_dict['revenue']
            )
            finance_dict['gross_profit_margin'] = StatsProcessor.cal_percentage(
                gross_profit_margin
            )
        except Exception as e:
            finance_dict['gross_profit_margin'] = None
            # print(f"gross_profit_margin failed because of {str(e)}")

    @classmethod
    def cal_operation_profit_rate(cls, finance_dict):
        """
        計算營業利益率
        營業利益率 = ( 營業收入－營業成本－營業費用）÷ 營業收入
        """
        try:
            operation_profit_rate = (
                finance_dict['revenue'] - finance_dict['operating_cost'] -
                finance_dict['operating_expenses']
            ) / finance_dict['revenue']
            finance_dict["operation_profit_rate"
                         ] = StatsProcessor.cal_percentage(
                             operation_profit_rate
                         )
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict["operation_profit_rate"] = None
            # print(f"operation_profit failed because of {str(e)}")

    @classmethod
    def cal_operating_cash_flow_profit_rate(cls, finance_dict):
        """
        計算營業現金流利潤率
        營業現金流利潤率 = 營業活動現金流 ÷ 營業收入
        """
        try:
            operating_cash_flow_profit_rate = (
                finance_dict["operating_cash_flow"] / finance_dict["revenue"]
            )
            finance_dict["operating_cash_flow_profit_rate"
                         ] = StatsProcessor.cal_percentage(
                             operating_cash_flow_profit_rate
                         )
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict["operating_cash_flow_profit_rate"] = None
            # print(
            #     f"operating_cash_flow_profit_rate failed because of {str(e)}")


# 成長動能

    """
    前四個已經有了 revenue_YoY, gross_prof_YoY, operating_income_YoY, net_income_YoY:
    後四個在金流，還需要處理
    """
    # 營運指標

    @classmethod
    def cal_dso(cls, finance_dict):
        """
        計算應收帳款收現天數(DSO)
        DSO = 365 × (應收帳款平均餘額 ÷ 營業收入)
        """
        try:
            dso = (
                365 * (finance_dict['account_pay'] / finance_dict['revenue'])
            )
            finance_dict['dso'] = StatsProcessor.cal_non_percentage(
                dso, to_str=True, postfix="日"
            )
        except Exception as e:
            finance_dict['dso'] = None
            # print(f"Error calculating 應收帳款收現天數 because of {str(e)}")

    @classmethod
    def cal_account_receive_over_revenue(cls, finance_dict):
        """
        計算應收帳款佔營收比率
        = 應收帳款平均餘額 ÷ 營業收入
        """
        try:
            account_receive_over_revenue = (
                finance_dict['account_receive'] / finance_dict['revenue']
            )
            finance_dict["account_receive_over_revenue"
                         ] = StatsProcessor.cal_percentage(
                             account_receive_over_revenue
                         )
        except Exception as e:
            finance_dict["account_receive_over_revenue"] = None

    @classmethod
    def cal_dpo(cls, finance_dict):
        """
        計算應付帳款週轉天數
        DPO = 365天 ÷ (銷貨成本÷平均應付帳款)
        """
        try:
            dpo = (
                365 *
                (finance_dict['account_pay'] / finance_dict['operating_cost'])
            )
            finance_dict["dpo"] = StatsProcessor.cal_non_percentage(
                dpo, to_str=True, postfix="日"
            )
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict["dpo"] = None
            # print(f"應付帳款週轉天數 failed because of {str(e)}")

    @classmethod
    def cal_inventories_cycle_ratio(cls, finance_dict):
        """
        計算存貨周轉率
        = 銷貨成本 ÷ 存貨
        """
        try:
            inventories_cycle_ratio = (
                finance_dict['operating_cost'] / finance_dict['inventories']
            )

            finance_dict["inventories_cycle_ratio"
                         ] = StatsProcessor.cal_percentage(
                             inventories_cycle_ratio
                         )
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict["inventories_cycle_ratio"] = None
            # print(f"Error calculating 存貨周轉率 because of {str(e)}")

    @classmethod
    def cal_dio(cls, finance_dict):
        """
        計算 存貨週轉天數 or 平均售貨天數
        DIO = 365天 * (存貨 ÷ 銷貨成本)
        MUDA MUDA MUDA !!!
        """
        try:
            dio = 365 * (
                finance_dict["inventories"] / finance_dict["operating_cost"]
            )
            finance_dict["dio"] = StatsProcessor.cal_non_percentage(
                dio, to_str=True, postfix="日"
            )
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict["dio"] = None
            # print(f"Error calculating 存貨週轉天數 because of {str(e)}")

    @classmethod
    def cal_inventories_revenue_ratio(cls, finance_dict):
        """
        計算存貨佔營收比率
        存貨佔營收比= 存貨 ÷ 營業收入
        """
        try:
            inventories_revenue_ratio = (
                finance_dict['inventories'] / finance_dict['revenue']
            )

            finance_dict["inventories_revenue_ratio"
                         ] = StatsProcessor.cal_percentage(
                             inventories_revenue_ratio
                         )
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict["inventories_revenue_ratio"] = None
            # print(f"Error calculating 存貨佔營收比率 because of {str(e)}")

    @classmethod
    def cal_cash_of_conversion_cycle(cls, finance_dict):
        """
        計算現金循環週期
        存貨週轉天數 + 應收帳款週轉天數 - 應付帳款週轉天數
        """
        try:
            cash_of_conversion_cycle = (
                finance_dict["dio"] + finance_dict["dso"] - finance_dict['dpo']
            )
            finance_dict["cash_of_conversion_cycle"
                         ] = StatsProcessor.cal_non_percentage(
                             cash_of_conversion_cycle, to_str=True, postfix="日"
                         )
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict["cash_of_conversion_cycle"] = None

    @classmethod
    def cal_asset_turnover(cls, finance_dict):
        """
        計算資產周轉率
        營業收入 ÷ 資產總額
        """
        try:
            asset_turnover = (
                finance_dict["revenue"] / finance_dict["inventories"]
            )
            finance_dict["asset_turnover"] = StatsProcessor.cal_percentage(
                asset_turnover
            )
        except Exception as e:
            finance_dict["asset_turnover"] = None

    @classmethod
    def cal_application_turnover(cls, finance_dict):
        """
        不動產、廠房及設備週轉率
        營業收入 ÷ 不動產、廠房與設備平均餘額
        """
        try:
            applcation_turnover = (
                finance_dict['revenue'] / finance_dict["application"]
            )
            finance_dict['application_turnover'
                         ] = StatsProcessor.cal_percentage(applcation_turnover)

        except Exception as e:
            finance_dict['application_turnover'] = None

    @classmethod
    def cal_current_ratio(cls, finance_dict):
        """
        計算流動比率 = 流動資產 / 流動負債
        """
        try:
            current_ratio = (
                finance_dict['current_assets'] /
                finance_dict['current_liabilities']
            )
            finance_dict['current_ratio'] = StatsProcessor.cal_percentage(
                current_ratio
            )
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['current_ratio'] = None
            # print(f"Error calculating current ratio: {e}")

    @classmethod
    def cal_quick_ratio(cls, finance_dict):
        """
        速動比率
        (流動資產 - 存貨) / 流動負債
        """
        try:
            quick_ratio = (
                finance_dict['current_assets'] - finance_dict['inventories']
            ) / finance_dict['current_liabilities']
            finance_dict['quick_ratio'] = StatsProcessor.cal_percentage(
                quick_ratio
            )
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['quick_ratio'] = None
            # print(f"Error calculating quick ratio: {e}")

    @classmethod
    def cal_debt_to_equity_ratio(cls, finance_dict):
        """
        # 負債權益比率 = 總負債 / 股東權益
        """
        try:
            debt_to_equity_ratio = finance_dict['total_liabilities'
                                                ] / finance_dict['equity']
            finance_dict['debt_to_equity_ratio'
                         ] = StatsProcessor.cal_percentage(
                             debt_to_equity_ratio
                         )
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['debt_to_equity_ratio'] = None
            # print(f"Error calculating debt to equity ratio: {e}")

    @classmethod
    def cal_net_debt_to_equity_ratio(cls, finance_dict):
        """
        # 淨負債權益比率 = (總負債 - 現金及約當現金) / 股東權益
        """
        try:
            net_debt_to_equity_ratio = (
                finance_dict['total_liabilities'] -
                finance_dict['cash_and_cash_equivalents']
            ) / finance_dict['equity']
            finance_dict['net_debt_to_equity_ratio'
                         ] = StatsProcessor.cal_percentage(
                             net_debt_to_equity_ratio
                         )
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['net_debt_to_equity_ratio'] = None
            # print(f"Error calculating net debt to equity ratio: {e}")

    @classmethod
    def cal_interest_coverage_ratio(cls, finance_dict):
        """
        # 利息保障倍數 = EBIT / 利息費用
        """
        try:
            interest_coverage_ratio = finance_dict['EBIT'] / finance_dict[
                'interest_expense']
            finance_dict['interest_coverage_ratio'
                         ] = StatsProcessor.cal_non_percentage(
                             interest_coverage_ratio, to_str=True, postfix="倍"
                         )
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['interest_coverage_ratio'] = None
            # print(f"Error calculating interest coverage ratio: {e}")

    @classmethod
    def cal_debt_to_operating_cash_flow(cls, finance_dict):
        """
        有息負債 / 營業活動現金流
        """
        try:
            debt_to_operating_cash_flow = finance_dict[
                'interest_bearing_debt'] / finance_dict['operating_cash_flow']
            finance_dict['debt_to_operating_cash_flow'
                         ] = StatsProcessor.cal_percentage(
                             debt_to_operating_cash_flow
                         )
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['debt_to_operating_cash_flow'] = None
            # print(f"Error calculating debt to operating cash flow: {e}")

    @classmethod
    def cal_debt_to_free_cash_flow(cls, finance_dict):
        """
        # 有息負債 / 自由現金流
        """
        try:
            debt_to_free_cash_flow = finance_dict['interest_bearing_debt'
                                                  ] / finance_dict['fcf']
            finance_dict['debt_to_free_cash_flow'
                         ] = StatsProcessor.cal_percentage(
                             debt_to_free_cash_flow
                         )
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['debt_to_free_cash_flow'] = None
            # print(f"Error calculating debt to free cash flow: {e}")

    @classmethod
    def cal_cash_flow_ratio(cls, finance_dict):
        """
        # 現金流量比率 = 營業活動現金流 / 流動負債
        """
        try:
            cash_flow_ratio = finance_dict[
                'operating_cash_flow'] / finance_dict['current_liabilities']
            finance_dict['cash_flow_ratio'] = StatsProcessor.cal_percentage(
                cash_flow_ratio
            )
        except (KeyError, ZeroDivisionError, TypeError) as e:
            finance_dict['cash_flow_ratio'] = None
            # print(f"Error calculating cash flow ratio: {e}")



from .base import BaseFinanceOverviewTransformer, FinanceOverviewProcessor
from neurostats_API.utils import StatsProcessor


class FinanceOverviewTransformer(BaseFinanceOverviewTransformer):

    def __init__(self, ticker, company_name, zone):
        super().__init__(ticker, company_name, zone)

        self.target_fields = StatsProcessor.load_yaml(
            "twse/finance_overview_dict.yaml"
        )
        self.inverse_dict = StatsProcessor.load_txt(
            "twse/seasonal_data_field_dict.txt", json_load=True
        )

    def process_transform(self, balance_sheet, profit_lose, cash_flow):

        processed_data_dict = self._filter_target_data(
            balance_sheet, profit_lose, cash_flow
        )
        FinanceOverviewProcessor.process_rate(processed_data_dict) # 處理百分比
        FinanceOverviewProcessor.process_all(processed_data_dict) # 計算所有指標公式
        self._fill_nan_index(processed_data_dict) # 處理空值
        FinanceOverviewProcessor.process_thousand_dollar(processed_data_dict) # 處理千元單位

        return processed_data_dict

    def _filter_target_data(self, balance_sheet, profit_lose, cash_flow):
        balance_sheet = balance_sheet[-1]
        cash_flow = cash_flow[-1]
        profit_lose = profit_lose[-1]
        seasons = [
            (int(report['year']), int(report['season']))
            for report in (balance_sheet, profit_lose, cash_flow)
        ]

        max_date = max(seasons)
        year, season = max_date

        data_dict = {
            'balance_sheet': balance_sheet['balance_sheet'],
            'profit_lose': profit_lose['profit_lose'],
            "cash_flow": cash_flow['cash_flow']
        }
        filtered_dict = {
            "year": year,
            "season": season
        }


        for key, target_sets in self.target_fields.items():
            try:
                small_targets = target_sets['field']
                value_index = target_sets['value']

                for small_target in small_targets:
                    big_target = self.inverse_dict[small_target]
                    # 透過inverse_dict取得項目在balance_sheet/profit_lose/cash_flow

                if (small_target == "利息費用_bank"):
                    small_target = small_target[:small_target.find("_bank")]

                seasonal_data = data_dict.get(big_target, {})    # 對應的列表

                filtered_dict[key] = seasonal_data.get(small_target, {}).get(
                    value_index, None
                )    # 取對應的值

            except Exception as e:
                print(f"filter failed :{e}")
                continue

        return filtered_dict

    def _fill_nan_index(self, finance_dict):
        for key in self.target_fields.keys():
            if (key not in finance_dict.keys()):
                finance_dict[key] = None



from .base import BaseTEJTransformer
import pandas as pd

class TEJChipTransformer(BaseTEJTransformer):
    def __init__(self, ticker, company_name, zone):
        super().__init__(ticker, company_name, zone)
        self.return_keys = ["chip_data"]
        

    def process_transform(
        self,
        fetched_data
    ):
        if (not fetched_data):
            return self._get_empty_structure()

        fetched_data = pd.DataFrame(fetched_data)

        return {
            "ticker": self.ticker,
            "company_name": self.company_name,
            "chip_data": fetched_data
        }



from .base import BaseCashFlowTransformer
from neurostats_API.utils import StatsProcessor, YoY_Calculator
import pandas as pd

class USCashFlowTransformer(BaseCashFlowTransformer):
    def __init__(self, ticker, company_name, zone):
        super().__init__(ticker, company_name, zone)
        self.data_df = None
        self.return_keys = ['cash_flow', 'cash_flow_YoY']
    def process_transform(self, fetched_data):
        if (not fetched_data):
            return self._get_empty_structure()

        self.data_df = self._process_us_format(fetched_data)
        target_season = fetched_data[0]['season']

        return_dict = {
            "ticker": self.ticker,
            "company_name": self.company_name,
            "cash_flow": self.data_df,
            "cash_flow_YoY": self._slice_target_season(self.data_df, target_season)
        }

        return return_dict
    
    def process_QoQ(self, fetched_data):
        if (not fetched_data):
            return self._get_empty_structure()
        
        data_df = self._process_twse_to_tej_format(fetched_data)
        target_season = fetched_data[0]['season']

        data_df_YoY = self._slice_target_season(data_df, target_season)

        return {
            "cash_flow_all": data_df,
            "cash_flow_YoY": data_df_YoY
        }


import dotenv
import json
import os
from pprint import PrettyPrinter
import pytest
import pandas as pd
import yaml
from pymongo import AsyncMongoClient

dotenv.load_dotenv(".env")
mongo_uri = os.getenv('MONGO_URI')
tickers = ["AAPL", "AMZN", "2330", "1260", "-1234"]

client = AsyncMongoClient(mongo_uri)

end_date = '2024-09-01'

pp = PrettyPrinter(indent=1)


pd.set_option("display.max_rows", 2)         # 顯示所有列
pd.set_option("display.max_columns", 2)      # 顯示所有欄
pd.set_option("display.width", 0)               # 自動根據 terminal 寬度換行
pd.set_option("display.max_colwidth", None)     # 顯示所有欄位字串長度（不省略）
pd.set_option("expand_frame_repr", False)


@pytest.mark.asyncio
async def test_balance_sheet():
    print("# Balance Sheet")
    from neurostats_API.async_mode.fetchers import AsyncBalanceSheetFetcher

    for ticker in tickers:
        fetcher = AsyncBalanceSheetFetcher(ticker, client)

        data = await fetcher.query_data()

        print_dict_summary(data)
    print('========================================')



@pytest.mark.asyncio
async def test_profit_lose():
    print("# Profit Lose")
    from neurostats_API.async_mode.fetchers import AsyncProfitLoseFetcher

    for ticker in tickers:
        fetcher = AsyncProfitLoseFetcher(ticker, client)

        data = await fetcher.query_data(start_date='2024-01-01' ,end_date=end_date)

        print_dict_summary(data)
    print('========================================')
    


@pytest.mark.asyncio
async def test_cash_flow():
    print("# Cash Flow")
    from neurostats_API.async_mode.fetchers import AsyncCashFlowFetcher
    for ticker in tickers:
        fetcher = AsyncCashFlowFetcher(ticker, client)

        data = await fetcher.query_data(end_date=end_date)

        print_dict_summary(data)
    print('========================================')
    


@pytest.mark.asyncio
async def test_cash_flow_partial():
    print("# Cash Flow (QoQ and YoY only)")
    from neurostats_API.async_mode.fetchers import AsyncCashFlowFetcher
    for ticker in tickers:
        fetcher = AsyncCashFlowFetcher(ticker, client)

        data = await fetcher.query_data_QoQ_and_YoY_only(end_date=end_date)

        print_dict_summary(data)
    print('========================================')


@pytest.mark.asyncio
async def test_month_revenue():
    print("# Month Revenue")
    from neurostats_API.async_mode.fetchers import AsyncMonthlyRevenueFetcher
    for ticker in tickers:
        fetcher = AsyncMonthlyRevenueFetcher(ticker, client)

        data = await fetcher.query_data(end_date=end_date)

        print_dict_summary(data)
    print('========================================')


@pytest.mark.asyncio
async def test_tech():
    print("# Tech")
    from neurostats_API.async_mode.fetchers import AsyncTechFetcher

    for ticker in tickers:
        fetcher = AsyncTechFetcher(ticker, client)

        test_fns = [
            fetcher.get_daily, fetcher.get_weekly, fetcher.get_monthly,
            fetcher.get_quarterly, fetcher.get_yearly
        ]

        for fetch_fn in test_fns:
            print(f"{fetch_fn.__name__}")
            data = await fetch_fn()

            print(data)
    print('========================================')
@pytest.mark.asyncio
async def test_institution():
    print("# Institution trading")
    from neurostats_API.async_mode.fetchers import AsyncTWSEInstitutionFetcher

    for ticker in tickers:
        fetcher = AsyncTWSEInstitutionFetcher(ticker, client)

        test_fns = [fetcher.query_data]

        for fetch_fn in test_fns:
            data = await fetch_fn()

            print_dict_summary(data)
    print('========================================')


@pytest.mark.asyncio
async def test_margin():
    print("# Margin trading")
    from neurostats_API.async_mode.fetchers import AsyncTWSEMarginFetcher

    for ticker in tickers:
        fetcher = AsyncTWSEMarginFetcher(ticker, client)

        test_fns = [fetcher.query_data, fetcher.query_data_annual]

        for fetch_fn in test_fns:
            data = await fetch_fn()

            print_dict_summary(data)
    print('========================================')

@pytest.mark.asyncio
async def test_value():
    print("# Valuation")
    from neurostats_API.async_mode.fetchers import AsyncTWSEStatsValueFetcher

    for ticker in tickers:
        fetcher = AsyncTWSEStatsValueFetcher(ticker, client)

        test_fn = [fetcher.query_data]

        for fetch_fn in test_fn:
            data = await fetch_fn()

            print_dict_summary(data)
    print('========================================')

@pytest.mark.asyncio
@pytest.mark.parametrize("ticker", tickers)
@pytest.mark.parametrize("collection", ["TEJ_finance_statement"])
@pytest.mark.parametrize("params", [
    {"fetch_mode": "QoQ", "use_cal": False, "indexes": []},
    {"fetch_mode": "QoQ", "use_cal": True, "indexes": []},
    {"fetch_mode": "QoQ", "use_cal": False, "indexes": ["eps"]},
    {"fetch_mode": "QoQ", "use_cal": True, "indexes": ["eps"]},
    {"fetch_mode": "YoY", "use_cal": False, "indexes": []},
    {"fetch_mode": "YoY", "use_cal": True, "indexes": []},
    {"fetch_mode": "YoY", "use_cal": False, "indexes": ["eps"]},
    {"fetch_mode": "YoY", "use_cal": True, "indexes": ["eps"]},
])
async def test_tej_fetchers(ticker, collection, params):
    from neurostats_API.utils import NoDataError
    from neurostats_API.async_mode.fetchers import AsyncTEJSeasonalFetcher
    print(f"# TEJ [{collection}] {ticker} - {params}")
    fetcher = AsyncTEJSeasonalFetcher(
        ticker=ticker,
        client=client,
        collection=collection
    )

    try:
        data = await fetcher.query_data(**params)
        
        print(data)
    except NoDataError:
        print(f"!! No data for [{collection}] {ticker} !!")
    print('========================================')


@pytest.mark.asyncio
async def test_stats_overview():
    print(f"# stats overview")
    from neurostats_API.async_mode.fetchers import AsyncFinanceOverviewFetcher

    for ticker in tickers:
        fetcher = AsyncFinanceOverviewFetcher(ticker, client)

        test_fn = [fetcher.query_data]

        for fetch_fn in test_fn:
            data = await fetch_fn()

            print_dict_summary(data)
    print('========================================')

@pytest.mark.asyncio
async def test_agent_overview():
    print(f"# Agents overview")
    from neurostats_API.async_mode.fetchers import AsyncAgentOverviewFetcher

    for ticker in tickers:
        fetcher = AsyncAgentOverviewFetcher(ticker, client)

        test_fn = [fetcher.query_data]

        for fetch_fn in test_fn:
            data = await fetch_fn()

            print_dict_summary(data)
    print('========================================')

@pytest.mark.asyncio
async def test_us_chip():
    print(f"# US F13 Chip")
    from neurostats_API.async_mode.fetchers import AsyncUSChipF13Fetcher
    
    mongo_uri = "mongodb+srv://admin:Neurowatt456&@axonnews-mongodb.nlui1.mongodb.net/?retryWrites=true&w=majority&appName=axonnews-mongodb"
    client = AsyncMongoClient(mongo_uri)

    for ticker in tickers:
        fetcher = AsyncUSChipF13Fetcher(ticker, client, None)

        test_fn = [fetcher.query_data]

        for fetch_fn in test_fn:
            try:
                data = await fetch_fn(end_date=end_date)

                print(data)
            except Exception as e:
                print(e)
    print('========================================')

@pytest.mark.asyncio
async def test_tej_chip():
    print(f"# Agents overview")
    from neurostats_API.async_mode.fetchers import AsyncTEJChipFetcher

    for ticker in tickers:
        try:
            fetcher = AsyncTEJChipFetcher(ticker, client)

            test_fn = [fetcher.query_data]

            for fetch_fn in test_fn:
                data = await fetch_fn()

                print_dict_summary(data)
                print('========================================')
        
        except Exception as e:
            print(str(e))

def print_dict_summary(data: dict, indent: int = 1):
    prefix = " " * indent
    for key, value in data.items():
        if isinstance(value, pd.DataFrame):
            assert value.shape!= (0, 0)
            print(f"{prefix} \"{key}\": DataFrame shape = {value.shape}")
        elif isinstance(value, pd.Series):
            print(f"{prefix} \"{key}\": Series shape = {value.shape}")
        elif isinstance(value, dict):
            print(f"{prefix} \"{key}\": dict with {len(value)} keys")
            print_dict_summary(value, indent + 2)
        elif isinstance(value, list):
            print(f"{prefix} \"{key}\": list of length {len(value)}")
        else:
            val_preview = str(value)
            if len(val_preview) > 100:
                val_preview = val_preview[:100] + "..."
            print(f"{prefix} \"{key}\": {type(value).__name__} = {val_preview}")


from .base import AsyncBaseSeasonalDBExtractor
from datetime import datetime
from neurostats_API.utils import NotSupportedError
from pymongo import ASCENDING


class AsyncTEJSeasonalDBExtractor(AsyncBaseSeasonalDBExtractor):

    def __init__(self, ticker, client, fetch_type='Q'):
        super().__init__(ticker, client)

        self.fetch_type = fetch_type

        if (self.zone != 'tw'):
            raise NotSupportedError(None)

    async def query_data(self, start_date=None, end_date=None):
        if (start_date is None):
            start_date = "1991-01-01"
        if (end_date is None):
            end_date = self._get_today()

        query, projection, sort = self._prepare_query(start_date, end_date)

        cursor = self.collection.find(query, projection).sort(sort)

        fetched_data = [data async for data in cursor]

        return fetched_data

    def _prepare_query(self, start_date, end_date):
        start_date = self._transform_date(start_date)
        end_date = self._transform_date(end_date)

        query = {
            "ticker": self.ticker,
            "date": {
                "$gte": start_date,
                "$lte": end_date
            }
        }

        projection = {
            "_id": 0,
            "year": 1,
            "season": 1,
            self.fetch_type: 1,
        }

        sorting = [("year", ASCENDING), ("season", ASCENDING)]

        return query, projection, sorting

    def _transform_date(self, date):
        if (isinstance(date, str)):
            try:
                return datetime.strptime(date, "%Y-%m-%d")
            except Exception as e:
                raise ValueError(
                    "date string format imcompatible, should be \"YYYY-MM-DD\""
                )
        elif (isinstance(date, datetime)):
            return date
        else:
            raise ValueError(
                "date type imcompatible, should be string(\"YYYY-MM-DD\") or datetime.datetime"
            )


class AsyncTEJFinanceStatementDBExtractor(AsyncTEJSeasonalDBExtractor):

    def __init__(self, ticker, mongo_uri, fetch_type='Q'):
        super().__init__(ticker, mongo_uri, fetch_type=fetch_type)

    def _get_collection_map(self):
        return {
            'tw': "TEJ_finance_statement"
        }
    
class AsyncTEJSelfSettlementDBExtractor(AsyncTEJSeasonalDBExtractor):
    def __init__(self, ticker, mongo_uri, fetch_type='Q'):
        super().__init__(ticker, mongo_uri, fetch_type=fetch_type)

    def _get_collection_map(self):
        return {
            'tw': "TEJ_self_settlement"
        }


from  ..base import BaseTransformer
import pandas as pd

class BaseProfitLoseTransformer(BaseTransformer):
    def __init__(self, ticker, company_name, zone):
        super().__init__(ticker, company_name, zone)

    def _process_twse_to_stats_format(self, fetched_data):
        table_dict = {}
        for data in fetched_data:
            year, season, profit_lose = data['year'], data['season'], data['profit_lose']

            time_index = f"{year}Q{season}"
            table_dict[time_index] = profit_lose

        stats_df = pd.DataFrame(table_dict)

        return stats_df
    
    def _process_twse_to_tej_format(self, fetched_data):
        """
        new_df 迎合 TEJ格式, 用於 report_generator
        """
        if (isinstance(fetched_data, list)):
            table_dict = {}
            for data in fetched_data:
                year, season, profit_lose = data['year'], data['season'], data['profit_lose']
                time_index = f"{year}Q{season}"

                new_profit_lose = self.flatten_dict(
                    profit_lose,
                    target_keys=[
                        "value", "growth", 
                        "percentage", "grand_total_percentage"
                    ] + [f"YoY_{i}" for i in [1, 3, 5, 10]]
                    + [f"grand_total_YoY_{i}" for i in [1, 3, 5, 10]]
                )

                table_dict[time_index] = new_profit_lose
            
            stats_df = pd.DataFrame.from_dict(table_dict)
            return stats_df.T
        
        elif (isinstance(fetched_data, dict)):
            for time_index, profit_lose in fetched_data.items():
                fetched_data[time_index] = self.flatten_dict(
                    value_dict=profit_lose,
                    target_keys=[
                        "value", "growth", "percentage", "grand_total",
                        "grand_total_percentage"
                    ] + [f"YoY_{i}" for i in [1, 3, 5, 10]] +
                    [f"grand_total_YoY_{i}" for i in [1, 3, 5, 10]]
                )

            stats_df = pd.DataFrame.from_dict(fetched_data)
            return stats_df.T
        
    def _process_us_format(self, fetched_data):
        """
        主要用於report generator
        """
        if (isinstance(fetched_data, list)):
            table_dict = {}
            for data in fetched_data:
                year, season, income_statement = data['year'], data['season'], data['income_statement']
                time_index = f"{year}Q{season}"

                table_dict[time_index] = income_statement
            
            stats_df = pd.DataFrame.from_dict(table_dict)
            return stats_df.T

        elif (isinstance(fetched_data, dict)):
            for time_index, income_statement in fetched_data.items():
                fetched_data[time_index] = self.flatten_dict(
                    value_dict=income_statement,
                    target_keys=["value", "growth"] +
                    [f"YoY_{i}" for i in [1, 3, 5, 10]]
                )
            stats_df = pd.DataFrame.from_dict(fetched_data)
            return stats_df.T




from typing import Type, TypedDict, Optional
from neurostats_API.async_mode.db_extractors import (
    AsyncTEJDailyTechDBExtractor, AsyncYFDailyTechDBExtractor,
    AsyncDailyValueDBExtractor, AsyncTEJDailyValueDBExtractor,
    AsyncTEJDailyChipDBExtractor, AsyncTWSEChipDBExtractor,
    AsyncTWSEMonthlyRevenueExtractor, AsyncBalanceSheetExtractor,
    AsyncProfitLoseExtractor, AsyncCashFlowExtractor,
    AsyncTEJFinanceStatementDBExtractor, AsyncTEJSelfSettlementDBExtractor,
    AsyncUS_F13DBExtractor
)
from neurostats_API.utils import NotSupportedError


class ExtractorConfig(TypedDict):
    class_: Type
    description: Optional[str]
    default_kwargs: dict


EXTRACTOR_MAP: dict[str, ExtractorConfig] = {
    # TEJ 相關
    "TEJ_finance_statement": {
        "class_":
        AsyncTEJFinanceStatementDBExtractor,
        "default_kwargs": {
            "fetch_type": "Q"
        },
        "description":
        """
        DB_collection: TEJ_finance_statement
        對應TEJ的\"會計師簽證財務資料\"
        """,
    },
    "TEJ_self_settlement": {
        "class_":
        AsyncTEJSelfSettlementDBExtractor,
        "default_kwargs": {
            "fetch_type": "Q"
        },
        "description":
        """
        DB_collection: TEJ_self_settlement
        對應TEJ的\"公司自結數\"
        """
    },

    # TWSE 本地DB
    "DB_balance_sheet": {
        "class_":
        AsyncBalanceSheetExtractor,
        "default_kwargs": {},
        "description":
        """
        DB_collection: twse_seasonal_report(tw), us_fundamental(us)
        對應TWSE的\"資產負債表\"與美股的\"Balance_Sheet\"
        tw: 選取balance_sheet欄
        us: 選取balance_sheet欄
        """
    },
    "DB_profit_lose": {
        "class_":
        AsyncProfitLoseExtractor,
        "default_kwargs": {},
        "description":
        """
        DB_collection: twse_seasonal_report(tw), us_fundamental(us)
        對應TWSE的\"損益表\"與美股的\"Income Statement\"
        tw: 選取profit_lose欄
        us: 選取income_statement欄
        """
    },
    "DB_cash_flow": {
        "class_":
        AsyncCashFlowExtractor,
        "default_kwargs": {},
        "description":
        """
        DB_collection: twse_seasonal_report(tw), us_fundamental(us)
        對應TWSE的\"金流表\"與美股的\"Cash Flow\"
        tw: 選取cash_flow欄
        us: 選取cash_flow欄
        """
    },

    # tech
    "YF_tech": {
        "class_":
        AsyncYFDailyTechDBExtractor,
        "default_kwargs": {},
        "description":
        """
        DB_collection: twse_daily_share_price(tw), us_tech(us)
        開高低收
        """
    },
    "TEJ_Chip": {
        "class_":
        AsyncTEJDailyChipDBExtractor,
        "default_kwargs": {},
        "description":
        """
        DB_collection: TEJ_chip
        TWSE的評價面
        """
    },
    "TEJ_tech": {
        "class_":
        AsyncTEJDailyTechDBExtractor,
        "default_kwargs": {},
        "description":
        """
        DB_collection: TEJ_share_price
        TEJ的開高低收
        """
    },

    # Month Revenue
    "TWSE_month_revenue": {
        "class_":
        AsyncTWSEMonthlyRevenueExtractor,
        "default_kwargs": {},
        "description":
        """
        DB_collection: twse_month_revenue
        TWSE的月營收
        """
    },

    # Chip
    "TWSE_chip": {
        "class_":
        AsyncTWSEChipDBExtractor,
        "default_kwargs": {
            "fetch_type": "I"
        },
        "description":
        """
        DB_collection: twse_chip
        TWSE的籌碼面
        """
    },

    # Value
    "TWSE_value": {
        "class_":
        AsyncDailyValueDBExtractor,
        "default_kwargs": {},
        "description":
        """
        DB_collection: twse_daily_share_price
        TWSE的評價面
        """
    },
    "US_Chip_F13": {
        "class_":
        AsyncUS_F13DBExtractor,
        "default_kwargs": {},
        "description":
        """
        DB_collection: "us_F13_flat"
        US籌碼面(F13)
        """
    }
}


def get_extractor(collection: str, ticker: str, client, **override_kwargs):
    """
    collection: extractor對應的KEY_NAME,
    ticker: 初始化extractor用的參數, 指定哪家公司
    client: 初始化extractor用的參數, 連接的db_client
    overide_kwargs: 除ticker與client外部分extractor會需要的參數，例如fetch_type
    """

    config = EXTRACTOR_MAP.get(collection, None)

    if (config is None):
        # 不存在的extractor name處理
        valid = ", ".join(EXTRACTOR_MAP.keys())
        raise NotSupportedError(
            f"{collection} not a regular argument for collection, only {[valid]} available"
        )

    extractor_cls = config["class_"]
    default_kwargs: dict = config["default_kwargs"]

    init_kwargs = {**default_kwargs, **override_kwargs}    # 用後續的參數覆蓋

    return extractor_cls(ticker, client, **init_kwargs)



from  ..base import BaseTransformer
import pandas as pd

class BaseBalanceSheetTransformer(BaseTransformer):
    def __init__(self, ticker, company_name, zone):
        super().__init__(ticker, company_name, zone)

    def _process_twse_to_stats_format(self, fetched_data):
        table_dict = {}
        for data in fetched_data:
            year, season, balance_sheet = data['year'], data['season'], data['balance_sheet']

            time_index = f"{year}Q{season}"
            table_dict[time_index] = balance_sheet

        stats_df = pd.DataFrame(table_dict)

        return stats_df
    
    def _process_twse_to_tej_format(self, fetched_data):
        """
        new_df 迎合 TEJ格式, 用於 report_generator
        """
        table_dict = {}
        for data in fetched_data:
            year, season, balance_sheet = data['year'], data['season'], data['balance_sheet']
            time_index = f"{year}Q{season}"

            new_balance_sheet = self.flatten_dict(
                balance_sheet,
                target_keys=["value", "percentage"]
            )

            table_dict[time_index] = new_balance_sheet
        
        stats_df = pd.DataFrame.from_dict(table_dict)
        return stats_df.T
    
    def _process_us_format(self, fetched_data):
        """
        主要用於report generator
        """
        table_dict = {}
        for data in fetched_data:
            year, season, balance_sheet = data['year'], data['season'], data['balance_sheet']
            time_index = f"{year}Q{season}"

            table_dict[time_index] = balance_sheet
        
        stats_df = pd.DataFrame.from_dict(table_dict)
        return stats_df.T



from .base import BaseDBClient
import os
from pymongo import AsyncMongoClient
from dotenv import load_dotenv

class USDBClient(BaseDBClient):
    def __init__(self, client):
        """初始化時接收 MongoDB 連接 URI"""
        super().__init__(client, zone = 'us')


from .base import BaseValueTransformer
import pandas as pd

class TWSEValueTransformer(BaseValueTransformer):
    def __init__(self, ticker, company_name, zone):
        super().__init__(ticker, company_name, zone)
    
    def process_transform(self, fetched_data):
        pass

class TWSEAnnualValueTransformer(TWSEValueTransformer):
    def __init__(self, ticker, company_name, zone):
        super().__init__(ticker, company_name, zone)
    
    def process_transform(self, annual_data):
        annual_df = pd.DataFrame(annual_data)

        return annual_df

class TWSEHistoryValueTransformer(TWSEValueTransformer):
    def __init__(self, ticker, company_name, zone):
        super().__init__(ticker, company_name, zone)
        self.column = [
            "P_E", "P_FCF", "P_B", "P_S",
            "EV_OPI", "EV_EBIT", "EV_EBITDA", "EV_S", "close"
        ]
    
    def process_transform(self, daily_data):
        history_df = pd.DataFrame(daily_data).set_index("date")
        return_dict = {}

        for key in self.column:
            return_dict[key] = history_df.loc[:, [key]]
        
        return return_dict
    
    def process_latest(self, daily_data):
        if (not daily_data):
            return self._get_empty_structure()

        daily_data = daily_data[-1]

        return daily_data
    
    def _get_empty_structure(self):
        return {
            key: None for key in self.column
        }


from .base import AsyncBaseFetcher
from datetime import datetime
import importlib.resources as pkg_resources

from neurostats_API.async_mode.db_extractors import (
    AsyncBalanceSheetExtractor, AsyncCashFlowExtractor, AsyncProfitLoseExtractor
)
from neurostats_API.async_mode.factory import get_extractor
from neurostats_API.transformers import AgentOverviewTransformer, FinanceOverviewTransformer
from neurostats_API.utils import StatsDateTime, StatsProcessor, NoCompanyError, NotSupportedError
import numpy as np
import pandas as pd
import pytz
import holidays
import warnings

import yaml


class AsyncFinanceOverviewFetcher(AsyncBaseFetcher):
    """
    對應iFa.ai -> 財務分析 -> 重要指標(finance_overview)
    """

    def __init__(self, ticker, db_client):
        super().__init__()

        self.target_fields = StatsProcessor.load_yaml(
            "twse/finance_overview_dict.yaml"
        )
        self.inverse_dict = StatsProcessor.load_txt(
            "twse/seasonal_data_field_dict.txt", json_load=True
        )
        try:
            self.balance_sheet_extractor = get_extractor(
                "DB_balance_sheet", ticker, db_client
            )
            self.profit_lose_extractor = get_extractor(
                "DB_profit_lose", ticker, db_client
            )
            self.cashflow_extractor = get_extractor(
                "DB_cash_flow", ticker, db_client
            )

            company_name = self.balance_sheet_extractor.get_company_name()
            zone = self.balance_sheet_extractor.get_zone()

            if (zone != 'tw'):
                raise NotSupportedError(None)

            self.transformer = FinanceOverviewTransformer(
                ticker, company_name, zone
            )
        except NotSupportedError as e:
            raise NotSupportedError("FinanceOverviewFetcher only supports TW corporation now") from e

    async def query_data(self, date=None):

        balance_sheet = await self.balance_sheet_extractor.query_data(
            end_date=date, get_latest=True
        )
        profit_lose = await self.profit_lose_extractor.query_data(
            end_date=date, get_latest=True
        )
        cash_flow = await self.cashflow_extractor.query_data(
            end_date=date, get_latest=True
        )

        transformed_data = self.transformer.process_transform(
            balance_sheet, profit_lose, cash_flow
        )

        return transformed_data


class AsyncAgentOverviewFetcher(AsyncBaseFetcher):
    """
    用於P1-Agent 的 公司股市經濟概況頁面
    """

    def __init__(self, ticker, db_client):
        from neurostats_API.async_mode.fetchers import AsyncTechFetcher
        self.ticker = ticker
        self.tech_fetcher = AsyncTechFetcher(
            ticker=self.ticker, client=db_client
        )
        try:
            self.balance_sheet_extractor = AsyncBalanceSheetExtractor(
                ticker=self.ticker, client=db_client
            )
            self.profit_lose_extractor = AsyncProfitLoseExtractor(
                ticker=self.ticker, client=db_client
            )
            self.cash_flow_extractor = AsyncCashFlowExtractor(
                ticker=self.ticker, client=db_client
            )

            company_name = self.balance_sheet_extractor.get_company_name()
            zone = self.balance_sheet_extractor.get_zone()

            self.transformer_map = {'us': AgentOverviewTransformer}

            transformer = self.get_transformer(zone)
            self.transformer = transformer(ticker, company_name, zone)

        except NotSupportedError as e:
            raise NotSupportedError(
                f'Agent Overview only supports US corporation now, Got zone={zone}, ticker={ticker}'
            ) from e

    async def query_data(self, date=None):

        tech_data = await self.tech_fetcher.get_daily(
            start_date=None, end_date=date
        )
        balance_sheet = await self.balance_sheet_extractor.query_data(
            end_date=date
        )
        profit_lose = await self.profit_lose_extractor.query_data(end_date=date)
        cash_flow = await self.cash_flow_extractor.query_data(end_date=date)

        return_data = self.transformer.process_transform(
            tech_data=tech_data,
            balance_sheet_list=balance_sheet,
            profit_lose_list=profit_lose,
            cash_flow_list=cash_flow
        )

        return return_data



from .base import AsyncBaseFetcher
from ..db_extractors import (
    AsyncDailyValueDBExtractor, AsyncTEJDailyValueDBExtractor
)
from neurostats_API.async_mode.factory import get_extractor
from neurostats_API.transformers import TWSEAnnualValueTransformer, TWSEHistoryValueTransformer
from neurostats_API.utils import NotSupportedError


class AsyncTWSEStatsValueFetcher(AsyncBaseFetcher):

    def __init__(self, ticker, client):

        self.ticker = ticker
        self.transformer_map = {
                "tw": (TWSEHistoryValueTransformer, TWSEAnnualValueTransformer)
            }
        try:
            self.daily_data_extractor = get_extractor("TWSE_value", ticker, client, fetch_type = 'D')
            self.annual_data_extractor = get_extractor("TWSE_value", ticker, client, fetch_type = 'Y')

            self.company_name = self.daily_data_extractor.get_company_name()
            self.zone = self.daily_data_extractor.get_zone()

            transformer = self.get_transformer(self.zone)

            daily_transformer, annual_transformer = transformer
            self.daily_transformer = daily_transformer(
                ticker, self.company_name, self.zone
            )
            self.annual_transformer = annual_transformer(
                ticker, self.company_name, self.zone
            )

            self.return_dict = {
                "ticker": self.ticker,
                "company_name": self.company_name
            }
        except NotSupportedError as e:
            raise NotSupportedError(
                f"{self.__class__.__name__} only support {list(self.transformer_map.keys())} companies now, {ticker} is not available"
            ) from e

    async def query_data(self, start_date=None, end_date=None):

        annual_data = await self.annual_data_extractor.query_data(
            start_date, end_date
        )
        daily_data = await self.daily_data_extractor.query_data(
            start_date=start_date, end_date=end_date
        )

        self.return_dict['yearly_data'
                         ] = self.annual_transformer.process_transform(
                             annual_data
                         )
        self.return_dict['daily_data'] = self.daily_transformer.process_latest(
            daily_data
        )
        self.return_dict['value_serie'
                         ] = self.daily_transformer.process_transform(
                             daily_data
                         )

        return self.return_dict

    async def query_value_serie(self, start_date=None, end_date=None):
        daily_data = await self.daily_data_extractor.query_data(
            start_date, end_date
        )
        self.return_dict['value_serie'
                         ] = self.daily_transformer.process_transform(
                             daily_data
                         )

        return self.return_dict



from .base import StatsFetcher, StatsDateTime
import importlib.resources as pkg_resources
import json
import numpy as np
import pandas as pd
from ..utils import StatsDateTime, StatsProcessor, YoY_Calculator
import yaml


class ProfitLoseFetcher(StatsFetcher):
    """
    iFa.ai: 財務分析 -> 損益表
    """

    def __init__(self, ticker, db_client):
        super().__init__(ticker, db_client)

        self.table_settings = StatsProcessor.load_yaml("twse/profit_lose.yaml")

        self.process_function_map = {
            "twse_stats": self.process_data_twse,
            "us_stats": self.process_data_us
        }

        self.return_keys = [
            'profit_lose', 'grand_total_profit_lose', 'revenue', 'grand_total_revenue', 
            'gross_profit', 'grand_total_gross_profit', 'gross_profit_percentage', 
            'grand_total_gross_profit_percentage', 'operating_income', 'grand_total_operating_income', 'operating_income_percentage', 
            'grand_total_operating_income_percentage', 'net_income_before_tax', 'grand_total_net_income_before_tax', 'net_income_before_tax_percentage', 
            'grand_total_net_income_before_tax_percentage', 'net_income', 'grand_total_net_income', 'net_income_percentage', 
            'grand_total_income_percentage', 'EPS', 'EPS_growth', 'grand_total_EPS', 
            'grand_total_EPS_growth', 'profit_lose_all', 'profit_lose_YoY'
        ]

    def prepare_query(self):
        pipeline = super().prepare_query()

        name_map = {"twse_stats": "profit_lose", "us_stats": "income_statement"}

        chart_name = name_map.get(self.collection_name, "income_statement")

        append_pipeline = [
        {
            "$project": {
                "_id": 0,
                "ticker": 1,
                "company_name": 1,
                "seasonal_data": {
                    "$map": {
                        "input": {"$ifNull": ["$seasonal_data", []]},
                        "as": "season",
                        "in": {
                            "year": "$$season.year",
                            "season": "$$season.season",
                            "data": {"$ifNull": [f"$$season.{chart_name}", []]}
                        }
                    }
                }
            }
        }
    ]

        pipeline = pipeline + append_pipeline

        return pipeline

    def collect_data(self):
        return super().collect_data()

    def query_data(self):
        fetched_data = self.collect_data()

        fetched_data = fetched_data[0]

        process_fn = self.process_function_map.get(
            self.collection_name, self.process_data_us
        )
        return process_fn(fetched_data)

    def process_data_twse(self, fetched_data):

        latest_time = StatsDateTime.get_latest_time(
            self.ticker, self.collection
        ).get('last_update_time', {})

        # 取最新時間資料時間，沒取到就預設去年年底
        target_year = latest_time.get('seasonal_data', {}).get(
            'latest_target_year',
            StatsDateTime.get_today().year - 1
        )
        target_season = latest_time.get('seasonal_data',
                                        {}).get('latest_season', 4)

        return_dict = {
            "ticker": self.ticker,
            "company_name": fetched_data['company_name'],
        }

        seasonal_data = fetched_data.get('seasonal_data', [])

        if (not seasonal_data):
            return_dict.update(self._get_empty_structure())
            return return_dict

        profit_lose_dict = {
            f"{data['year']}Q{data['season']}": data['data']
            for data in seasonal_data
        }

        profit_lose_dict = YoY_Calculator.cal_QoQ(profit_lose_dict)
        profit_lose_df = pd.DataFrame.from_dict(profit_lose_dict)
        target_season_col = profit_lose_df.columns.str.endswith(
            f"Q{target_season}"
        )
        profit_lose_df = profit_lose_df.loc[:, target_season_col]

        old_profit_lose_df = StatsProcessor.expand_value_percentage(
            profit_lose_df
        )
        # OLD: 回傳包含value & percentage
        value_col = old_profit_lose_df.columns.str.endswith(f"_value")
        percentage_col = old_profit_lose_df.columns.str.endswith(f"_percentage")
        # OLD: 回傳剔除grand_total
        grand_total_value_col = old_profit_lose_df.columns.str.endswith(
            f"grand_total_value"
        )
        grand_total_percentage_col = old_profit_lose_df.columns.str.endswith(
            f"grand_total_percentage"
        )

        old_profit_lose_df = old_profit_lose_df.loc[:, (
            (value_col & ~grand_total_value_col) |
            (percentage_col & ~grand_total_percentage_col)
        )]

        for time_index, data_dict in profit_lose_dict.items():
            profit_lose_dict[time_index] = self.flatten_dict(
                value_dict=data_dict,
                indexes=list(data_dict.keys()),
                target_keys=[
                    "value", "growth", "percentage", "grand_total",
                    "grand_total_percentage"
                ] + [f"YoY_{i}" for i in [1, 3, 5, 10]] +
                [f"grand_total_YoY_{i}" for i in [1, 3, 5, 10]]
            )

        profit_lose_df = pd.DataFrame.from_dict(profit_lose_dict).T
        # EPS的value用元計算
        eps_index = (
            profit_lose_df.columns.str.endswith("_value")
            & profit_lose_df.columns.str.contains("每股盈餘")
        )
        eps_copy = profit_lose_df.loc[:, eps_index].copy()
        eps_mask_index = eps_copy.columns
        profit_lose_df[eps_mask_index] = profit_lose_df[eps_mask_index].map(
            lambda x: StatsProcessor.cal_non_percentage(x, postfix="元")
        )

        # percentage處理
        percentage_index = profit_lose_df.columns.str.endswith("percentage")
        growth_index = profit_lose_df.columns.str.endswith("growth")
        percentage_mask = (percentage_index | growth_index)
        percentage_copy = profit_lose_df.loc[:, percentage_mask]
        percentage_mask_index = percentage_copy.columns

        profit_lose_df[percentage_mask_index] = profit_lose_df[
            percentage_mask_index].map(
                lambda x: StatsProcessor.
                cal_non_percentage(x, to_str=True, postfix="%")
            )

        # YoY處理: 乘以100
        YoY_index = profit_lose_df.columns.str.contains("YoY")
        YoY_mask = YoY_index
        YoY_copy = profit_lose_df.loc[:, YoY_mask]
        YoY_mask_cols = YoY_copy.columns

        profit_lose_df[YoY_mask_cols] = profit_lose_df[YoY_mask_cols].map(
            lambda x: StatsProcessor.cal_percentage(x)
        )

        # 剩下的處理: 乘以千元
        value_index = ~(
            percentage_index | growth_index | YoY_index | eps_index
        )    # 除了上述以外的 index

        value_col = profit_lose_df.loc[:, value_index].columns
        profit_lose_df[value_col] = profit_lose_df[value_col].map(
            lambda x: StatsProcessor.cal_non_percentage(x, postfix="千元")
        )

        total_table = profit_lose_df.replace("N/A", None).T

        # 取特定季度
        target_season_columns = total_table.columns.str.endswith(
            f"Q{target_season}"
        )
        total_table_YoY = total_table.loc[:, target_season_columns]

        for name, setting in self.table_settings.items():
            target_indexes = setting.get('target_index', [None])
            for target_index in target_indexes:
                try:
                    return_dict[name] = StatsProcessor.slice_table(
                        total_table=total_table_YoY,
                        mode=setting['mode'],
                        target_index=target_index
                    )
                    break
                except Exception as e:
                    continue

        return_dict.update(
            {
                "profit_lose": old_profit_lose_df,
                "profit_lose_all": total_table.copy(),
                "profit_lose_YoY": total_table_YoY
            }
        )
        return return_dict

    def process_data_us(self, fetched_data):

        table_dict = {
            f"{data['year']}Q{data['season']}": data['data']
            for data in fetched_data['seasonal_data']
        }

        table_dict = YoY_Calculator.cal_QoQ(table_dict)
        table_dict = YoY_Calculator.cal_YoY(table_dict)

        for time_index, data_dict in table_dict.items():
            table_dict[time_index] = self.flatten_dict(
                value_dict=data_dict,
                indexes=list(data_dict.keys()),
                target_keys=["value", "growth"] +
                [f"YoY_{i}" for i in [1, 3, 5, 10]]
            )

        # 計算QoQ

        return_dict = {
            "ticker": self.ticker,
            "company_name": fetched_data['company_name'],
            "profit_lose": pd.DataFrame.from_dict(table_dict)
        }

        return return_dict

    def _get_empty_structure(self):
        return {
            key: pd.DataFrame(columns= pd.Index([], name = 'date')) for key in self.return_keys
        }


import abc
from datetime import datetime
from ..db import TWSEDBClient, USDBClient
import json
import pytz
from neurostats_API.utils import StatsDateTime, StatsProcessor, YoY_Calculator, NoCompanyError

class BaseDBExtractor(abc.ABC):
    def __init__(self, ticker, client):
        self.ticker = ticker
        self.client = client
        self.timezone = pytz.timezone("Asia/Taipei")
        self.tw_company_list = StatsProcessor.load_json("company_list/tw.json")
        self.us_company_list = StatsProcessor.load_json("company_list/us.json")

        self.company_name, self.zone = self._set_company_name(self.ticker)

        self.db_client = self._set_db_connection()

    def _set_company_name(self, ticker):

        index_dict = {
            "^TWII": ("台灣加權指數", 'tw'),
            "^GSPC": ("S&P 500", 'us'),
            "^DJI": ("Dow Jones Indices", 'us'),
            "^IXIC": ("NASDAQ Composite", 'us')
        }

        for key, values in index_dict.items():
            if (key == ticker):
                company_name, zone  = values
                return company_name, zone

        company_lists = [
            (self.tw_company_list, 'tw'),
            (self.us_company_list, 'us')
        ]

        for company_list, zone in company_lists:
            company_name = company_list.get(ticker, None)
            if (company_name):
                return company_name, zone
        # 沒找到公司名稱

        raise NoCompanyError(f"Ticker-{ticker} not found in any company lists")
    
    def _set_db_connection(self):
        db_map = {
            'tw': TWSEDBClient,
            'us': USDBClient
        }
        db_client = db_map.get(self.zone, TWSEDBClient)
        return db_client(self.client)

    
    async def query_data(self):
        """
        user使用的接口
        """
        return NotImplementedError()

    @abc.abstractmethod
    def _prepare_query(self):
        pass
    
    def get_company_name(self):
        """
        回傳公司
        """
        return self.company_name
    
    def get_zone(self):
        """
        公司區域(TW, US)
        """
        return self.zone
    
    def _get_today(self):
        return datetime.today()


from .base import BaseProfitLoseTransformer
from neurostats_API.utils import StatsProcessor, YoY_Calculator
import pandas as pd

class USProfitLoseTransformer(BaseProfitLoseTransformer):
    def __init__(self, ticker, company_name, zone):
        super().__init__(ticker, company_name, zone)
        self.data_df = None
        self.return_keys = ['profit_lose']
    
    def process_transform(self, fetched_data):
        
        data_dict = self._process_us_format(fetched_data).T.to_dict()

        data_dict = YoY_Calculator.cal_QoQ(data_dict)
        data_dict = YoY_Calculator.cal_YoY(data_dict)

        self.data_df = self._process_us_format(data_dict)

        return_dict = {
            "ticker": self.ticker,
            "company_name": self.company_name,
            "profit_lose": self.data_df
        }

        return return_dict



# 本模組提供各種財報、評價、籌碼等資料轉換器的工廠函式，根據 zone（如 "tw"、"us"）自動選擇對應的 Transformer 類別。
# 若 zone 不支援則會拋出 NotSupportedError，方便統一管理與擴充。

from typing import Type, TypedDict, Optional
from neurostats_API.transformers import (
    AgentOverviewTransformer,
    TEJFinanceStatementTransformer,
    TWSEAnnualValueTransformer,
    TWSEBalanceSheetTransformer,
    TWSECashFlowTransformer,
    TWSEChipTransformer,
    TWSEHistoryValueTransformer,
    TWSEMonthlyRevenueTransformer,
    TWSEProfitLoseTransformer,
    USBalanceSheetTransformer,
    USCashFlowTransformer,
    USProfitLoseTransformer
)
from neurostats_API.utils import NotSupportedError


def get_balance_sheet_transformer(zone):
    """
    取得資產負債表（Balance Sheet）對應地區的 Transformer 類別。
    參數:
        zone: "tw"（台灣）或 "us"（美國）
    回傳:
        對應的 Transformer 類別
    """
    transformer_map = {
        "tw": TWSEBalanceSheetTransformer,
        "us": USBalanceSheetTransformer
    }
    return _get_transformer_by_zone("Balance Sheet", zone, transformer_map)

def get_cash_flow_transformer(zone):
    """
    取得現金流量表（Cash Flow）對應地區的 Transformer 類別。
    參數:
        zone: "tw" 或 "us"
    回傳:
        對應的 Transformer 類別
    """
    transformer_map = {
        "tw": TWSECashFlowTransformer,
        "us": USCashFlowTransformer
    }
    return _get_transformer_by_zone("Cash Flow", zone, transformer_map)

def get_profit_lose_transformer(zone):
    """
    取得損益表（Profit Lose, 又稱 Income Statement）對應地區的 Transformer 類別。
    參數:
        zone: "tw" 或 "us"
    回傳:
        對應的 Transformer 類別
    """

    transformer_map = {
        "tw": TWSEProfitLoseTransformer,
        "us": USProfitLoseTransformer
    }
    return _get_transformer_by_zone("Profit Lose (a.k.a. Income Statement)", zone, transformer_map)

def get_chip_transformer(zone):
    """
    取得籌碼面（Chip）對應地區的 Transformer 類別。
    參數:
        zone: "tw"
    回傳:
        對應的 Transformer 類別
    """
    transformer_map = {
        "tw": TWSEChipTransformer
    }
    return _get_transformer_by_zone("Chip", zone, transformer_map)


def get_monthly_revenue_transformer(zone):
    """
    取得月營收（Monthly Revenue）對應地區的 Transformer 類別。
    參數:
        zone: "tw"
    回傳:
        對應的 Transformer 類別
    """
    transformer_map = {
        "tw": TWSEMonthlyRevenueTransformer
    }
    return _get_transformer_by_zone("MonthlyRevenue", zone, transformer_map)


def get_annual_value_transformer(zone):
    """
    取得歷年評價面數值（Annual Value）對應地區的 Transformer 類別。
    參數:
        zone: "tw"
    回傳:
        對應的 Transformer 類別
    """
    transformer_map = {
        "tw": TWSEAnnualValueTransformer
    }
    return _get_transformer_by_zone("AnnualValue", zone, transformer_map)


def get_history_value_transformer(zone):
    """
    取得歷來評價面數值（History Value）對應地區的 Transformer 類別。
    參數:
        zone: "tw"
    回傳:
        對應的 Transformer 類別
    """
    transformer_map = {
        "tw": TWSEHistoryValueTransformer
    }
    return _get_transformer_by_zone("HistoryValue", zone, transformer_map)


def get_agent_overview_transformer(zone):
    """
    取得美股券商總覽（Agent Overview）對應地區的 Transformer 類別。
    參數:
        zone: "us"
    回傳:
        對應的 Transformer 類別
    """
    transformer_map = {
        "us": AgentOverviewTransformer
    }
    return _get_transformer_by_zone("AgentOverview", zone, transformer_map)


def get_tej_finance_statement_transformer(zone):
    """
    取得 TEJ 財報（TEJ Finance Statement）對應地區的 Transformer 類別。
    參數:
        zone: "tw"
    回傳:
        對應的 Transformer 類別
    """
    transformer_map = {
        "tw": TEJFinanceStatementTransformer
    }
    return _get_transformer_by_zone("TEJFinanceStatement", zone, transformer_map)

# === 共用內部工具 ===
def _get_transformer_by_zone(name: str, zone: str, transformer_map: dict):
    """
    根據 zone 回傳對應的 Transformer 類別，若 zone 不支援則拋出 NotSupportedError。
    參數:
        name: 字串，Transformer 名稱（用於錯誤訊息）
        zone: 地區代碼
        transformer_map: dict，zone 對應的 Transformer 類別
    回傳:
        對應的 Transformer 類別
    """
    transformer = transformer_map.get(zone)
    if not transformer:
        raise NotSupportedError(
            f"[transformer_factory] {name} Transformer only supports {list(transformer_map.keys())}, got '{zone}'"
        )
    return transformer



from .tej_chip import AsyncTEJDailyChipDBExtractor
from .tej_tech import AsyncTEJDailyTechDBExtractor, AsyncBatchTechDBExtractor
from .twse_chip import AsyncTWSEChipDBExtractor
from .value import (
    AsyncDailyValueDBExtractor, 
    AsyncTEJDailyValueDBExtractor
)
from .us_chip import AsyncUS_F13DBExtractor, AsyncUS_CelebrityDBExtractor
from .yf import AsyncYFDailyTechDBExtractor


from .base import BaseCashFlowTransformer
from neurostats_API.utils import StatsProcessor

class TWSECashFlowTransformer(BaseCashFlowTransformer):
    def __init__(self, ticker, company_name, zone):
        super().__init__(ticker, company_name, zone)

        self.cash_flow_dict = StatsProcessor.load_yaml(
            "twse/cash_flow_percentage.yaml"
        )  # 計算子表格用

        self.return_keys = ['cash_flow', 'CASHO', 'CASHI', 'CASHF', 'cash_flow_all', 'cash_flow_YoY']

    def process_transform(self, fetched_data):
        """
        轉換整體
        """
        if (not fetched_data):
            return self._get_empty_structure()

        return_dict = {
            "ticker": self.ticker,
            "company_name": self.company_name
        }

        return_dict.update(self.process_stats_page(fetched_data))
        return_dict.update(self.process_QoQ(fetched_data))

        return return_dict
    def process_stats_page(self, fetched_data):
        """
        這裡主要回傳stats頁面相關的部分
        """
        if (not fetched_data):
            return self._get_empty_structure()

        stats_page_df, CASHO_df, CASHI_df, CASHF_df = self._process_twse_to_stats_format(fetched_data)
        target_season = fetched_data[0]['season']

        return {
            'cash_flow': self._slice_and_expand(stats_page_df, target_season),
            "CASHO": self._slice_and_expand(CASHO_df, target_season),
            "CASHI": self._slice_and_expand(CASHI_df, target_season),
            "CASHF": self._slice_and_expand(CASHF_df, target_season)
        }
    
    def process_QoQ(self, fetched_data):
        """
        這裡主要只會回傳其他地方需要的部分，此符合TEJ格式
        """
        if (not fetched_data):
            return self._get_empty_structure()
        
        data_df = self._process_twse_to_tej_format(fetched_data)
        data_df = self._cal_QoQ(data_df.T.to_dict())
        data_df = data_df.T
        target_season = fetched_data[0]['season']

        data_df_YoY = self._slice_target_season(data_df, target_season)

        return {
            "cash_flow_all": data_df,
            "cash_flow_YoY": data_df_YoY
        }

    def _slice_and_expand(self, data_df, target_season):
        data_df = self._slice_target_season(data_df, target_season)
        data_df = StatsProcessor.expand_value_percentage(data_df)

        return data_df


from datetime import datetime, timedelta
from pymongo import ASCENDING, DESCENDING
from .base import BaseDailyTechDBExtractor
from neurostats_API.utils import (
    NotSupportedError, StatsProcessor
)
from pymongo import AsyncMongoClient

class AsyncUS_F13DBExtractor(BaseDailyTechDBExtractor):

    def __init__(self, ticker, client, managerTicker=None):
        """
        ticker: Issuer Ticker, 被持有的股票代碼
        managerTicker: 持有者的股票代碼
        """
        super().__init__(ticker, client)
        self.issuerTicker = ticker
        self.managerTicker = managerTicker

        if (self.get_zone() not in ['us']):
            raise NotSupportedError("Supports US Company Only")

    def _get_collection_name(self):
        self.collection_name_map = {"us": "us_F13"}

        return self.collection_name_map.get(self.zone, None)

    def _prepare_query(self, start_date=None, end_date=None, get_latest=False):
        query = {"issuerTicker": self.issuerTicker}

        if (self.managerTicker):
            query.update(
                {
                    "managerTicker": self.managerTicker
                }
            )

        query = self._update_query_with_date(query, start_date, end_date)

        projection = {
            "_id": 0,
        }

        if (get_latest):
            sort = [("reportDate", DESCENDING)]
        else:
            sort = [("reportDate", ASCENDING)]

        return query, projection, sort

    def _update_query_with_date(self, query, start_date, end_date):
        start_date = self._transform_date(start_date)
        end_date = self._transform_date(end_date)

        date_range = {"$gte": start_date, "$lte": end_date}

        query.update({"reportDate": date_range})

        return query

    async def query_data(
        self, start_date=None, end_date=None, get_latest=False
    ):
        result = await super().query_data(start_date, end_date, get_latest)

        return result

class AsyncUS_CelebrityDBExtractor(AsyncUS_F13DBExtractor):
    """        
    用於查詢美國名人持股的專用extractor
    """
    def __init__(self, client:AsyncMongoClient, manager_name=None):
        """
        manager_name: 名人名稱
        """
        self.client = client
        self.zone = "us"
        self.manager = manager_name
        self.collection_name_map = {"us": "us_F13"}
        self.collection = self.client.get_database("company_us").get_collection("us_F13")
    

    def _get_collection_name(self):

        return self.collection_name_map.get(self.zone, None)
    
    def _prepare_query(
            self, start_date=None, end_date=None, get_latest=False,
            value_throshold=None, title=None
        ):
        query = {"manager": self.manager}

        if start_date:
            start_date = self._transform_date(start_date)
        if end_date:
            end_date = self._transform_date(end_date)

        if start_date or end_date:
            date_range = {}
            if start_date:
                date_range["$gte"] = start_date
            if end_date:
                date_range["$lte"] = end_date
            query["reportDate"] = date_range

        if value_throshold:
            query["value"] = {"$gt": value_throshold}

        if title:
            query["titleOfClass"] = title

        projection = {"_id": 0}

        if (get_latest):
            sort = [("reportDate", DESCENDING)]
        else:
            sort = [("reportDate", ASCENDING)]

        return query, projection, sort
    
    async def query_data(
        self, start_date=None, end_date=None, get_latest=False,
        value_throshold=None, title=None
    ):
        query, projection, sort = self._prepare_query(
            start_date=start_date, end_date=end_date, get_latest=get_latest,
            value_throshold=value_throshold, title=title
        )

        if (get_latest):
            cursor = self.collection.find(query, projection).sort(sort).limit(1)
        else:
            cursor = self.collection.find(query, projection).sort(sort)

        fetched_data = [data async for data in cursor]

        return fetched_data
    


from pymongo import AsyncMongoClient

class BaseDBClient:
    def __init__(self, client:  AsyncMongoClient, zone= None):
        self.client = client
        self.zone = zone or "tw"  # 預設台灣
        self.db = self._select_db()

    def _select_db(self):
        """
        根據 zone 選擇對應的資料庫
        """
        zone_db_map = {
            "tw": "company_test",
            "us": "company_us"
        }
        db_name = zone_db_map.get(self.zone, "company_test")
        return self.client[db_name]

    def get_collection(self, collection_name: str):
        """
        傳回指定 collection
        """
        return self.db[collection_name]



from .base import BaseDailyTechDBExtractor
from pymongo import AsyncMongoClient, ASCENDING, DESCENDING

class AsyncTEJDailyTechDBExtractor(BaseDailyTechDBExtractor):

    def __init__(self, ticker, client):
        super().__init__(ticker, client)

    def _get_collection_name(self):
        self.collection_name_map = {"tw": "TEJ_share_price"}

        return self.collection_name_map.get(self.zone, None)

class AsyncBatchTechDBExtractor(BaseDailyTechDBExtractor):

    def __init__(self, ticker:list, client: AsyncMongoClient, zone:str):
        self.tickers = ticker
        self.zone = zone
        self.client = client

        self.collection_map = {
            'tw':{
                'db': "company_test",
                'collection': "twse_daily_share_price"
            },
            'us':{
                'db': "company_us",
                'collection': "us_technical"
            },            
        }

        db_name = self.collection_map.get(self.zone).get('db')
        collection_name = self.collection_map.get(self.zone).get('collection')
        self.collection = self.client.get_database(db_name).get_collection(collection_name)

    def _prepare_query(
        self, start_date=None, end_date=None, get_latest=False
    ):
        query = {"ticker":{"$in": self.tickers}}

        if start_date:
            start_date = self._transform_date(start_date)
        if end_date:
            end_date = self._transform_date(end_date)

        if start_date or end_date:
            date_range = {}
            if start_date:
                date_range["$gte"] = start_date
            if end_date:
                date_range["$lte"] = end_date
            query["date"] = date_range

        projection = {
            "_id": 0,
            "ticker": 1,
            "date": 1,
            "open": 1,
            "high": 1,
            "low": 1,
            "close": 1,
            "volume": 1
        }

        if (get_latest):
            sort = [("date", DESCENDING)]
        else:
            sort = [("date", ASCENDING)]

        return query, projection, sort
    
    async def query_data(
        self, start_date=None, end_date=None, get_latest=False
    ):
        query, projection, sort = self._prepare_query(
            start_date=start_date, end_date=end_date, get_latest=get_latest
        )

        if (get_latest):
            cursor = self.collection.find(query, projection).sort(sort).limit(1)
        else:
            cursor = self.collection.find(query, projection).sort(sort)

        fetched_data = [data async for data in cursor]

        return fetched_data


from .data_process import StatsProcessor
import numpy as np
class YoY_Calculator:
    def __init__(self):
        pass

    @classmethod
    def cal_growth(cls, target_value: float, past_value: float, delta: int):
        """
        計算成長率以及年化成長率
        target_value: float，這個時間的數值
        past_value: float，過去的這個時間數值
        delta: int，代表隔了幾年/季 delta > 1 時改以年化成長率計算
        """
        try:
            if (delta > 1):
                YoY = ((target_value / past_value)**(1 / delta)) - 1

            else:
                YoY = ((target_value - past_value) / past_value)

        except Exception as e:
            return None

        if (isinstance(YoY, complex)): # 年化成長率有複數問題
            return None
        
        if YoY != YoY:  # 確認是否為nan (Python 中 nan不等於自己)
            return None

        return YoY
    @classmethod
    def calculate_growth(cls, this_value, last_value, delta):
        try:
            return YoY_Calculator.cal_growth(
                this_value, last_value, delta
            ) * 100
        except Exception:
            return None

    @classmethod
    def cal_YoY(
        cls, data_dict: dict
    ):
        year_shifts = [1, 3, 5, 10]
        return_dict = {}

        for time_index in data_dict.keys():
            year, season = map(int, time_index.split("Q"))
            
            year_data = data_dict.get(f"{year}Q{season}", {}).copy()
            if not year_data:
                continue

            for key, data in list(year_data.items()):
                if key in ["year", "season"]:
                    continue
                
                if (isinstance(data, dict)):
                    temp_dict = data
                    value = data.get("value", None)
                else:
                    temp_dict = {"value": data}
                    value = data

                this_value = StatsProcessor.process_str_to_float(value)
                
                for shift in year_shifts:
                    past_value = data_dict.get(f"{year - shift}Q{season}", {}).get(key, {})
                    if (isinstance(past_value, dict)):
                        past_value = past_value.get('value', None)
                    past_value = StatsProcessor.process_str_to_float(
                        past_value
                    )

                    growth = cls.calculate_growth(this_value, past_value, shift) if past_value else None

                    temp_dict[
                        f"YoY_{shift}"
                    ] = f"{growth:.2f}%" if growth else None

                year_data[key] = temp_dict
            return_dict[f"{year}Q{season}"] = year_data
        return return_dict
    @classmethod
    def cal_QoQ(cls, data_dict):
        return_dict = {}

        for time_index, this_data in data_dict.items():
            year, season = map(int, time_index.split("Q"))
            last_year, last_season = (
                year - 1, 4
            ) if season == 1 else (year, season - 1)

            for key in list(this_data.keys()):
                if key == "season":
                    continue

                value = this_data.get(key, None)

                if (isinstance(value, dict)):
                    temp_dict = value
                    this_value = value.get("value", None)
                else:
                    temp_dict = {"value": value}
                    this_value = value
                
                this_value = StatsProcessor.process_str_to_float(this_value)
                
                last_value = data_dict.get(
                    f"{last_year}Q{last_season}",{}
                ).get(key, {})

                if (isinstance(last_value, dict)):
                    last_value = last_value.get("value", None)

                last_value = StatsProcessor.process_str_to_float(last_value)
                growth = cls.calculate_growth(
                    this_value, last_value, 1
                ) if last_value is not None else None
                temp_dict['growth'] = (f"{growth:.2f}%" if growth else None)

                this_data[key] = temp_dict

            return_dict[time_index] = this_data

        return return_dict



from ..base import BaseTransformer
from neurostats_API.utils import StatsProcessor
import numpy as np
import pandas as pd

class BaseCashFlowTransformer(BaseTransformer):
    def __init__(self, ticker, company_name, zone):
        super().__init__(ticker, company_name, zone)
    
    def _process_twse_to_stats_format(self, fetched_data):
        table_dict = {}
        CASHO_dict = {}
        CASHI_dict = {}
        CASHF_dict = {}

        checkpoints = ["營業活動之現金流量－間接法", "投資活動之現金流量", "籌資活動之現金流量", "匯率變動對現金及約當現金之影響"]
        main_cash_flow_names = [
            "營業活動之淨現金流入（流出）", "投資活動之淨現金流入（流出）", "籌資活動之淨現金流入（流出）", "其他"
        ] # 主要的比例對象
        partial_cash_flow_dicts = [CASHO_dict, CASHI_dict, CASHF_dict, dict()] 

        boundary = len(main_cash_flow_names) - 1
        
        for data in fetched_data:
            year, season, cash_flow = data['year'], data['season'], data['cash_flow']

            time_index = f"{year}Q{season}"
            table_dict[time_index] = cash_flow

            # 處理sub part
            main_cash_flow_name = ""
            partial_cash_flow = {}
            next_checkpoint = 0

            temp_dict = {}

            for index_name, cash_flow_value in cash_flow.items():
                if (next_checkpoint < boundary
                        and index_name == checkpoints[next_checkpoint]): # 找到了主要的變動點
                    main_cash_flow_name = main_cash_flow_names[next_checkpoint]
                    partial_cash_flow = partial_cash_flow_dicts[next_checkpoint]
                    partial_cash_flow[time_index] = {}
                    next_checkpoint += 1

                if main_cash_flow_name: # 有取得cash flow name再進行接下來的計算percentage
                    if (isinstance(cash_flow_value, dict)):
                        value = cash_flow_value.get('value', None) 
                    else:
                        value = cash_flow_value

                    # 處理cashflow percentage部分(取2位 + 接%)
                    main_value = cash_flow.get(main_cash_flow_name, None)
                    if (isinstance(main_value, dict)):
                        main_value = main_value.get('value', None)
                    else:
                        pass

                    try:
                        ratio = np.round(
                            (value / main_value) * 100, 2
                        )
                        ratio = f"{ratio}%"
                    except:
                        ratio = None
                    
                    value = StatsProcessor.cal_non_percentage(value, postfix="千元")
                    temp_dict[index_name] = {
                        "value" : value,
                        "percentage": ratio
                    }
                    partial_cash_flow[time_index][index_name] = temp_dict[index_name]

            table_dict[time_index] = temp_dict

        stats_df = pd.DataFrame(table_dict)

        CASHO_df = pd.DataFrame(CASHO_dict)
        CASHI_df = pd.DataFrame(CASHI_dict)
        CASHF_df = pd.DataFrame(CASHF_dict)

        return stats_df, CASHO_df, CASHI_df, CASHF_df
    
    def _process_twse_to_tej_format(self, fetched_data):
        """
        new_df 迎合 TEJ格式, 用於 report_generator
        """
        table_dict = {}
        for data in fetched_data:
            year, season, cash_flow = data['year'], data['season'], data['cash_flow']
            time_index = f"{year}Q{season}"

            new_cash_flow = self.flatten_dict(
                cash_flow,
                target_keys=["value", "percentage"]
            )

            table_dict[time_index] = new_cash_flow
        
        stats_df = pd.DataFrame.from_dict(table_dict)
        return stats_df
    
    def _process_us_format(self, fetched_data):
        """
        主要用於report generator
        """
        table_dict = {}
        for data in fetched_data:
            year, season, cash_flow = data['year'], data['season'], data['cash_flow']
            time_index = f"{year}Q{season}"

            table_dict[time_index] = cash_flow
        
        stats_df = pd.DataFrame.from_dict(table_dict)
        return stats_df



from ..base import BaseDBExtractor
from datetime import datetime
import json
import pandas as pd
from pymongo import ASCENDING, DESCENDING
from neurostats_API.async_mode.db import TWSEDBClient, USDBClient
from neurostats_API.utils import StatsDateTime, StatsProcessor, NotSupportedError
import yaml


class AsyncBaseMonthlyDBExtractor(BaseDBExtractor):

    def __init__(self, ticker, client):
        super().__init__(ticker, client)

        self.collection_name_map = None
        self.collection_name = self._get_collection_name()

        if (self.collection_name is None):
             raise NotSupportedError(
                    f"{self.__class__.__name__} only supports {list(self.collection_name_map.keys())}, got {self.zone} with ticker = \"{self.ticker}\""
                )

        self.collection = self.db_client.get_collection(self.collection_name)

    async def query_data(
        self, start_date=None, end_date=None, get_latest=False
    ):
        if (start_date is None):
            start_date = "1991-01-01"
        if (end_date is None):
            end_date = self._get_today()

        start_year, start_month = self._get_year_and_month(date=start_date)
        end_year, end_month = self._get_year_and_month(date=end_date)

        query, projection, sort = self._prepare_query(
            start_year, start_month, end_year, end_month, get_latest
        )
        cursor = self.collection.find(query, projection).sort(sort)
        if (get_latest):
            cursor = cursor.limit(1)

        fetched_data = [data async for data in cursor]

        return fetched_data

    def _get_collection_name(self):
        self.collection_name_map = {
            "tw": "twse_monthly_revenue",
        }

        collection_name = self.collection_name_map.get(self.zone, None)

        return collection_name

    def _prepare_query(
        self,
        start_year=None,
        start_month=None,
        end_year=None,
        end_month=None,
        get_latest=False
    ):

        query = {"ticker": self.ticker}

        query = self._update_query_with_year_month(
            query, start_year, start_month, end_year, end_month
        )

        projection = {
            "_id": 0,
        }

        if (get_latest):
            sort = [("year", DESCENDING), ("month", DESCENDING)]
        else:
            sort = [("year", ASCENDING), ("month", ASCENDING)]

        return query, projection, sort

    def _get_year_and_month(self, date):
        if (isinstance(date, str)):
            date = datetime.strptime(date, "%Y-%m-%d")

        year = date.year
        month = date.month

        return year, month

    def _update_query_with_year_month(
        self, query, start_year, start_month, end_year, end_month
    ):
        if all(v is not None for v in [start_year, start_month]):
            # 大於start_year條件
            query.update(
                {
                    "$or": [
                        {
                            "year": {
                                "$gt": start_year
                            }
                        },
                        {
                            "year": start_year,
                            "month": {
                                "$gte": start_month
                            }
                        },
                    ]
                }
            )

        if all(v is not None for v in [end_year, end_month]):
            # 小於end_year條件
            query.update(
                {
                    "$and":
                    query.get("$and", []) + [
                        {
                            "$or": [
                                {
                                    "year": {
                                        "$lt": end_year
                                    }
                                },
                                {
                                    "year": end_year,
                                    "month": {
                                        "$lte": end_month
                                    }
                                },
                            ]
                        }
                    ]
                }
            )

        return query



from .base import BaseChipTransformer
from datetime import date, datetime
import inspect
import numpy as np
import pandas as pd
from neurostats_API.utils import StatsProcessor


class TWSEChipTransformer(BaseChipTransformer):

    def __init__(self, ticker, company_name, zone):
        super().__init__(ticker, company_name, zone)

    def process_transform(self, tech_data, **fetched_datas):
        """
        fetched_datas:
            {
                "institution_trading": {}
                "margin_trading": {},
                "security_lending: {},
            }
        """
        if not fetched_datas:
            return self._get_empty_structure()    # 若查無資料，回傳空表結構

        return_dict = {
            "ticker": self.ticker,
            "company_name": self.company_name,
            'latest_trading': {},
            'annual_trading': {},
            'price': {}
        }

        for key, fetched_data in fetched_datas.items():

            tech_dict = {data['date']: data for data in tech_data}
            latest_tech = tech_data[-1]
            if (not fetched_data):
                return_dict['latest_trading'].update(
                    self._process_latest({}, latest_tech, key)
                )
                return_dict['annual_trading'].update(
                    self._process_annual({}, tech_dict, key)
                )
                continue

            fetched_dict = {data['date']: data[key] for data in fetched_data}
            latest_data = fetched_data[-1]

            return_dict['latest_trading'].update(
                self._process_latest(latest_data, latest_tech, key)
            )
            return_dict['annual_trading'].update(
                self._process_annual(fetched_dict, tech_dict, key)
            )

        tech_df = pd.DataFrame(tech_data)
        return_dict['price'] = self._process_tech(tech_df)

        return return_dict

    def _process_latest(self, latest_data: dict, latest_tech: dict, key: str):
        latest_trade = latest_data.get(key)
        return {
            "date":
            latest_data.get('date',
                            datetime.today().strftime("%Y-%m-%d")),
            **self._process_latest_trading(latest_trade, latest_tech, key)
        }

    def _process_annual(self, fetched_dict: dict, tech_dict: dict, key: str):
        if (fetched_dict):
            fetched_data = [
                {
                    'date': date,
                    'close': tech_dict.get(date, {}).get('close', 0.0),
                    'volume': tech_dict.get(date, {}).get('volume', 0.0),
                    **trading
                } for date, trading in fetched_dict.items()
            ]
            fetched_df = self._process_annual_trading(fetched_data, key)

        else:
            fetched_data = [
                {
                    'date': date,
                    'close': tech_dict.get(date, {}).get('close', 0.0),
                    'volume': tech_dict.get(date, {}).get('volume', 0.0),
                } for date in tech_dict.keys()
            ]
            fetched_df = pd.DataFrame(fetched_data)

        return {key: fetched_df}

    def _process_latest_trading(self, latest_trading, latest_tech, key):
        FN_MAP = {
            'institution_trading': self._process_latest_institution,
            'margin_trading': self._process_latest_margin,
            'security_lending': self._process_latest_security
        }

        process_fn = FN_MAP.get(key)
        return process_fn(latest_trading, latest_tech)

    def _process_annual_trading(self, fetched_data, key):
        FN_MAP = {
            'institution_trading': self._process_annual_institution,
            'margin_trading': self._process_annual_margin,
            'security_lending': self._process_annual_security
        }

        process_fn = FN_MAP.get(key)
        return process_fn(fetched_data)

    def _process_tech(self, tech_df):
        latest_daily_data = tech_df.iloc[-1]
        yesterday_daily_data = tech_df.iloc[-2]

        price_dict = {
            "open": round(latest_daily_data['open'], 2),
            'close': round(latest_daily_data['close'], 2),
            'range':
            f"{latest_daily_data['low']:.2f} - {latest_daily_data['high']:.2f}",
            'volume': round(latest_daily_data['volume'] / 1000, 2),
            'last_open': round(yesterday_daily_data['open'], 2),
            'last_close': round(yesterday_daily_data['close'], 2),
            'last_range':
            f"{yesterday_daily_data['low']:.2f} - {yesterday_daily_data['high']:.2f}",
            'last_volume': round(yesterday_daily_data['volume'] / 1000, 2)
        }

        lowest = pd.to_numeric(tech_df["low"], errors="coerce").min().item()
        highest = pd.to_numeric(tech_df["high"], errors="coerce").max().item()
        price_dict['52weeks_range'] = f"{lowest:.2f} - {highest:.2f}"

        return price_dict

    def _process_latest_institution(self, latest_trading, latest_tech):

        def _default_institution_chart():
            return {
                "buy": {
                    "stock": None,
                    "price": None,
                    "average_price": None,
                    "percentage": None
                },
                "sell": {
                    "stock": None,
                    "price": None,
                    "average_price": None,
                    "percentage": None
                },
                "over_buy_sell": {
                    "stock": None,
                    "price": None,
                    "average_price": None,
                    "percentage": None
                },
            }

        latest_table = {
            "foreign": _default_institution_chart(),
            "mutual": _default_institution_chart(),
            "prop": _default_institution_chart(),
            "institutional_investor": _default_institution_chart(),
        }

        volume = latest_tech['volume']

        if (latest_trading is not None):
            for key in latest_trading.keys():
                if (key.find("外陸資") >= 0 or key.find("外資") >= 0):
                    self._target_institution(
                        latest_trading, latest_table['foreign'], key, volume
                    )
                elif (key.find("自營商") >= 0):
                    self._target_institution(
                        latest_trading, latest_table['prop'], key, volume
                    )
                elif (key.find("投信") >= 0):
                    self._target_institution(
                        latest_trading, latest_table['mutual'], key, volume
                    )
                elif (key.find("三大法人") >= 0):
                    self._target_institution(
                        latest_trading, latest_table['institutional_investor'],
                        key, volume
                    )
            # 計算合計
            for unit in ['stock', 'percentage']:
                # 買進總和
                latest_table['institutional_investor']['buy'][unit] = (
                    latest_table['foreign']['buy'][unit] +
                    latest_table['prop']['buy'][unit] +
                    latest_table['mutual']['buy'][unit]
                )
                # 賣出總和
                latest_table['institutional_investor']['sell'][unit] = (
                    latest_table['foreign']['sell'][unit] +
                    latest_table['prop']['sell'][unit] +
                    latest_table['mutual']['sell'][unit]
                )

        frames = []
        for category, trades in latest_table.items():
            temp_df = pd.DataFrame(trades).T
            temp_df['category'] = category
            frames.append(temp_df)

        latest_df = pd.concat(frames)
        latest_df = latest_df.reset_index().rename(columns={'index': 'type'})
        latest_df = latest_df[[
            'type', 'category', 'stock', 'price', 'average_price', 'percentage'
        ]]

        latest_df = pd.melt(
            latest_df,
            id_vars=['type', 'category'],
            var_name='variable',
            value_name='value'
        )

        latest_df = latest_df.pivot_table(
            index=['category', 'variable'],
            columns='type',
            values='value',
            aggfunc='first'
        )

        # 重設列名，去除多層索引
        latest_df.columns.name = None    # 去除列名稱
        latest_df = latest_df.reset_index()

        return {"institution_trading": latest_df}

    def _process_annual_institution(self, trading_datas):

        return pd.DataFrame(trading_datas)

    def _process_annual_margin(self, trading_datas):
        # 融資
        def _process(data):
            if ('現金償還' in data['financing'].keys()):
                data['financing']['現償'] = data['financing'].pop('現金償還')
                data['short_selling']['現償'] = data['short_selling'].pop(
                    '現券償還', data['financing']['現償']
                )
            financing = {
                f"融資_{type_name}": num_of_stocks
                for type_name, num_of_stocks in data['financing'].items()
            }

            short_selling = {
                f"融券_{type_name}": num_of_stocks
                for type_name, num_of_stocks in data['short_selling'].items()
            }

            security_offset = data['資券互抵']

            return {**financing, **short_selling, '資券互抵': security_offset}

        if (trading_datas):
            return_datas = [
                {
                    'date': data['date'],
                    'close': data['close'],
                    'volume': data['volume'],
                    **_process(data)
                } for data in trading_datas
            ]
        else:
            return_datas = []

        # 接起來
        return_datas = pd.DataFrame(return_datas)

        return return_datas

    def _process_annual_security(self, trading_datas):

        def _process(data):
            stock_lendings = {
                f"借券_{type_name}":
                (num_of_stocks /
                 1000) if isinstance(num_of_stocks,
                                     (int, float)) else num_of_stocks
                for type_name, num_of_stocks in data['stock_lending'].items()
            }

            return stock_lendings

        if (trading_datas):
            return_datas = [
                {
                    'date': data['date'],
                    'close': data['close'],
                    'volume': data['volume'],
                    **_process(data)
                } for data in trading_datas
            ]
        else:
            return_datas = []

        # 接起來
        return_datas = pd.DataFrame(return_datas)

        return return_datas

    def _target_institution(self, old_table, new_table, key, volume):
        if (key.find("買進") >= 0):
            self._cal_institution(old_table, new_table['buy'], key, volume)
        elif (key.find("賣出") >= 0):
            self._cal_institution(old_table, new_table['sell'], key, volume)
        elif (key.find("買賣超") >= 0):
            self._cal_institution(
                old_table, new_table['over_buy_sell'], key, volume
            )

    def _cal_institution(self, old_table, new_table, key, volume):
        new_table['stock'] = np.round(old_table[key] / 1000, 2).item()
        new_table['percentage'] = np.round((old_table[key] / volume) * 100,
                                           2).item()

    def _process_latest_margin(self, latest_trading, *args):

        def _default_margin_chart():
            dafault_dict = {"financing": {}, "short_selling": {}}
            return {
                "margin_trading": pd.DataFrame.from_dict(dafault_dict),
                "security_offset": None
            }

        if (latest_trading is None):
            return _default_margin_chart()

        latest_trading['financing']['現償'] = latest_trading['financing'].pop(
            '現金償還'
        )
        latest_trading['short_selling']['現償'] = latest_trading['short_selling'
                                                               ].pop('現券償還')

        latest_trading_df = {
            category: sub_dict
            for category, sub_dict in latest_trading.items()
            if (isinstance(sub_dict, dict))
        }

        latest_margin_trading_df = pd.DataFrame.from_dict(latest_trading_df)

        return {
            "margin_trading": latest_margin_trading_df,
            "security_offset": latest_trading['資券互抵']
        }

    def _process_latest_security(self, latest_trading, *args):

        def _default_margin_chart():
            return {
                "stock_lending":
                pd.DataFrame(
                    index=["當日賣出", "現償", "當日還券", "當日調整", "當日餘額", "次一營業日可限額"]
                )
            }

        if (latest_trading is None):
            return _default_margin_chart()

        latest_stock_lending = latest_trading['stock_lending']

        latest_stock_lending = {
            type_name: StatsProcessor.cal_round_int(value / 1000)
            for type_name, value in latest_stock_lending.items()
        }
        latest_stock_lending.pop("前日餘額")
        latest_stock_lending_df = pd.DataFrame.from_dict(
            latest_stock_lending, orient="index", columns=['stock_lending']
        )

        return latest_stock_lending_df

    def _get_empty_structure(self):
        return_dict = {
            "ticker": self.ticker,
            "company_name": self.company_name,
        }

        return_dict['latest_trading'] = {
            'date': date.today(),
            'institution_trading': pd.DataFrame(),
            'margin_trading': pd.DataFrame(),
            'security_lending': pd.DataFrame()
        }

        return_dict['annual_trading'] = {
            'institution_trading':
            pd.DataFrame(columns=['date', 'close', 'volume']),
            'margin_trading': pd.DataFrame(columns=['date', 'close', 'volume']),
            'security_lending':
            pd.DataFrame(columns=['date', 'close', 'volume']),
        }

        return_dict['price'] = {
            "open": None,
            'close': None,
            'range': None,
            'volume': None,
            'last_open': None,
            'last_close': None,
            'last_range': None,
            'last_volume': None,
            '52weeks_range': None
        }

        return return_dict



__version__='1.0.3rc1'

from .fetchers import (
    AgentFinanceOverviewFetcher,
    BalanceSheetFetcher,
    CashFlowFetcher,
    FinanceOverviewFetcher,
    FinanceReportFetcher,
    InstitutionFetcher,
    MarginTradingFetcher,
    MonthRevenueFetcher,
    TechFetcher,
    TEJStockPriceFetcher,
    ProfitLoseFetcher,
)

from .async_mode import (
    AsyncAgentOverviewFetcher,
    AsyncBalanceSheetFetcher,
    AsyncCashFlowFetcher,
    AsyncFinanceOverviewFetcher,
    AsyncMonthlyRevenueFetcher,
    AsyncProfitLoseFetcher,
    AsyncTechFetcher,
    AsyncTEJSeasonalFetcher,
    AsyncTWSEInstitutionFetcher,
    AsyncTWSEMarginFetcher,
    AsyncTWSEStatsValueFetcher,
    AsyncUSCelebrityFetcher,
    AsyncBatchTechFetcher
)


from .base import BaseTEJTransformer
import pandas as pd


class TEJTechTransformer(BaseTEJTransformer):

    def __init__(self, ticker, company_name, zone):
        super().__init__(ticker, company_name, zone)
        self.fetched_data = None

    def process_transform(self, fetched_data):
        self.fetched_data = pd.DataFrame(fetched_data)

        self.fetched_data = self.fetched_data.rename(
            columns={
                "open_d": "open",
                "high_d": "high",
                "low_d": "low",
                "close_d": "close",
                "vol": "volume"
            }
        )

        return self.fetched_data



from ..base import BaseDBExtractor
from datetime import datetime, timezone
from neurostats_API.utils import NotSupportedError
from pymongo import ASCENDING, DESCENDING


class BaseDailyTechDBExtractor(BaseDBExtractor):

    def __init__(self, ticker, client):
        super().__init__(ticker, client)

        self.collection_name_map = None
        self.collection_name = self._get_collection_name()
        
        if (self.collection_name is None):
             raise NotSupportedError(
                    f"{self.__class__.__name__} only supports {list(self.collection_name_map.keys())}, got {self.zone} with ticker = \"{self.ticker}\""
                )

        self.collection = self.db_client.get_collection(self.collection_name)

    async def query_data(self, start_date=None, end_date=None, get_latest = False):
        if (self.collection is None):
            return []

        if (start_date is None):
            start_date = "1991-01-01"
        if (end_date is None):
            end_date = self._get_today()

        query, projection, sort = self._prepare_query(
            start_date=start_date, end_date=end_date, get_latest=get_latest
        )

        if (get_latest):
            cursor = self.collection.find(query, projection).sort(sort).limit(1)
        else:
            cursor = self.collection.find(query, projection).sort(sort)

        fetched_data = [data async for data in cursor]

        return fetched_data

    def _get_collection_name(self):
        self.collection_name_map = {"tw": "twse_daily_share_price", "us": "us_technical"}
        collection_name = self.collection_name_map.get(self.zone, None)

        return collection_name 

    def _prepare_query(self, start_date=None, end_date=None, get_latest = False):
        query = {"ticker": self.ticker}

        query = self._update_query_with_date(query, start_date, end_date)

        projection = {
            "_id": 0,
        }

        if (get_latest):
            sort = [("date", DESCENDING)]
        else:
            sort = [("date", ASCENDING)]

        return query, projection, sort

    def _update_query_with_date(self, query, start_date, end_date):
        start_date = self._transform_date(start_date)
        end_date = self._transform_date(end_date)

        date_range = {"$gte": start_date, "$lte": end_date}

        query.update({"date": date_range})

        return query

    def _transform_date(self, date, use_UTC = False):
        if (isinstance(date, str)):
            try:
                date = datetime.strptime(date, "%Y-%m-%d")
                if (use_UTC):
                    date = date.astimezone(tz = timezone.utc)
                return date 
            except Exception as e:
                print(e)
                raise ValueError(
                    "date string format imcompatible, should be \"YYYY-MM-DD\""
                )
        elif (isinstance(date, datetime)):
            if (use_UTC):
                return date.astimezone(tz = timezone.utc)
            else:
                return date
        else:
            raise ValueError(
                "date type imcompatible, should be string(\"YYYY-MM-DD\") or datetime.datetime"
            )


from .base import BaseBalanceSheetTransformer
from neurostats_API.utils import StatsProcessor


class TWSEBalanceSheetTransformer(BaseBalanceSheetTransformer):

    def __init__(self, ticker, company_name, zone):
        super().__init__(ticker, company_name, zone)

        self.table_settings = StatsProcessor.load_yaml(
            "twse/balance_sheet.yaml"
        )
        self.return_keys = [
            'balance_sheet', 'total_asset', 'current_asset',
            'non_current_asset', 'current_debt', 'non_current_debt', 'equity',
            'balance_sheet_all', 'balance_sheet_YoY'
        ]

        self.stats_df = None
        self.new_df = None

    def process_transform(self, fetched_data):
        if (not fetched_data):
            return self._get_empty_structure()

        processed_data = self._process_fn(fetched_data)

        return processed_data

    def _process_fn(self, fetched_data):

        return_dict = {"ticker": self.ticker, "company_name": self.company_name}

        target_season = fetched_data[0]['season']

        self.stats_df = self._process_twse_to_stats_format(fetched_data)
        self.stats_df = self._slice_target_season(self.stats_df, target_season)

        # 轉換格式 (元 / 千元 / %)
        self.stats_df = StatsProcessor.expand_value_percentage(self.stats_df)
        self.stats_df = self._apply_process_unit_pipeline(
            self.stats_df, postfix_list=['_value', "_percentage"]
        )

        self.new_df = self._process_twse_to_tej_format(fetched_data)
        # 轉換格式 (元 / 千元 / %)
        self.new_df = self._apply_process_unit_pipeline(
            self.new_df, postfix_list=['_value', "_percentage"]
        )

        self.new_dict = self.new_df.to_dict()
        self.new_df = self._cal_QoQ(self.new_dict)
        self.new_df = self.new_df.T

        total_table_YoY = self._slice_target_season(
            self.new_df, target_season
        )

        return_dict.update(
            {
                "balance_sheet": self.stats_df,
                "balance_sheet_all": self.new_df,
                "balance_sheet_YoY": total_table_YoY
            }
        )
        # 抽取出小欄位整合
        self._process_target_columns(return_dict, self.stats_df)

        return return_dict

    def _process_target_columns(self, return_dict, stats_df):
        for name, setting in self.table_settings.items():
            target_indexes = setting.get('target_index', [None])
            for target_index in target_indexes:
                try:
                    return_dict[name] = StatsProcessor.slice_old_table(
                        total_table=stats_df, target_index=target_index
                    )
                    break
                except Exception as e:
                    continue



import pandas as pd
from neurostats_API.async_mode.fetchers.base import AsyncBaseFetcher
from neurostats_API.async_mode.factory import get_extractor
from neurostats_API.transformers import TEJTechTransformer
from neurostats_API.utils import NotSupportedError


class AsyncTEJDailyTechFetcher(AsyncBaseFetcher):

    def __init__(
        self,
        ticker,
        client,
    ):
        self.ticker = ticker
        try:
            self.extractor = get_extractor("TEJ_tech", ticker, client)

            self.company_name = self.extractor.get_company_name()
            self.zone = self.extractor.get_zone()

            self.transformer = TEJTechTransformer(
                ticker, self.company_name, self.zone
            )

        except NotSupportedError as e:
            raise NotSupportedError(
                f"{self.__class__.__name__} only support TW companies now, {ticker} is not available"
            ) from e

    async def query_data(self, start_date, end_date):
        data = await self.extractor.query_data(start_date, end_date)
        df = pd.DataFrame(data)
        return df.rename(
            columns={
                "open_d": "open",
                "high_d": "high",
                "low_d": "low",
                "close_d": "close",
                "vol": "volume"
            }
        )



from .base import AsyncBaseFetcher
from ..db_extractors import (
    AsyncDailyValueDBExtractor, AsyncTEJDailyValueDBExtractor, AsyncUS_CelebrityDBExtractor
)
from neurostats_API.async_mode.factory import get_extractor
from neurostats_API.transformers import US_F13_Transformer, US_CelebrityTransformer
from neurostats_API.utils import NotSupportedError


class AsyncUSChipF13Fetcher(AsyncBaseFetcher):
    def __init__(self, 
        ticker, 
        client,
        managerTicker = None,
    ):
        self.issuerTicker = ticker
        self.managerTicker = managerTicker
        self.transformer_map = {"us": US_F13_Transformer}
        try:
            self.extractor = get_extractor(
                "US_Chip_F13", ticker, client, managerTicker=self.managerTicker
            )

            company_name = self.extractor.get_company_name()
            zone = self.extractor.get_zone()

            transformer = self.get_transformer(zone)
            self.transformer = transformer(ticker, company_name, zone)

        except NotSupportedError as e:
            raise NotSupportedError(
                f"{self.__class__.__name__} only support {list(self.transformer_map.keys())} companies now, {ticker} is not available"
            ) from e

    async def query_data(
        self, start_date=None, end_date=None, get_latest=False
    ):
        fetched_data = await self.extractor.query_data(
            start_date, end_date, get_latest
        )

        transformed_data = self.transformer.process_transform(
            fetched_data
        )

        return transformed_data

class AsyncUSCelebrityFetcher(AsyncBaseFetcher):

    def __init__(self, client, manager_name):
        self.manager_name = manager_name
        self.extractor = AsyncUS_CelebrityDBExtractor(client, manager_name)
        self.transformer = US_CelebrityTransformer()

    async def query_data(
        self, start_date=None, end_date=None, get_latest=False,
        value_threshold=None, title=None, strftime=None
    ):
        fetched_data = await self.extractor.query_data(
            start_date, end_date, get_latest,
            value_throshold=value_threshold, title=title
        )

        transformed_data = self.transformer.process_transform(
            fetched_datas = list(fetched_data),
            strftime = strftime
        )

        return transformed_data


from neurostats_API.async_mode.fetchers.base import AsyncBaseFetcher
from datetime import datetime
from neurostats_API.async_mode.db_extractors import (
    AsyncTEJFinanceStatementDBExtractor, AsyncTEJSelfSettlementDBExtractor
)
from neurostats_API.async_mode.factory import get_extractor
from neurostats_API.transformers import TEJFinanceStatementTransformer
from neurostats_API.utils import NotSupportedError


class AsyncTEJSeasonalFetcher(AsyncBaseFetcher):

    def __init__(
        self,
        ticker,
        client,
        collection='TEJ_finance_statement',
        fetch_type='Q'
    ):
        self.ticker = ticker
        self.fetch_type = fetch_type
        self.transformer_map = {"tw": TEJFinanceStatementTransformer}
        try:
            self.db_extractor = get_extractor(
                collection=collection,
                ticker=ticker,
                client=client,
                fetch_type=fetch_type
            )

            self.company_name = self.db_extractor.get_company_name()
            self.zone = self.db_extractor.get_zone()

            

            transformer = self.get_transformer(self.zone)

            self.transformer = transformer(
                ticker=ticker,
                company_name=self.company_name,
                zone=self.zone,
                target_column=fetch_type
            )

        except NotSupportedError as e:
            raise NotSupportedError(
                f"{self.__class__.__name__} only support {list(self.transformer_map.keys())} companies now, {ticker} is not available"
            ) from e

    async def query_data(
        self,
        start_date=None,
        end_date=None,
        fetch_mode='QoQ',
        use_cal=True,
        indexes=None
    ):
        fetched_data = await self.db_extractor.query_data(
            start_date=None, end_date=end_date
        )    # 先索取目標日期之前的之後再截斷

        if (start_date is None):
            start_date = datetime.strptime("1991-01-01", "%Y-%m-%d")

        transformed_data = self.transformer.process_transform(
            fetched_data=fetched_data,
            fetch_mode=fetch_mode,
            start_date=start_date,
            use_cal=use_cal,
            indexes=indexes
        )    # 轉換並根據開始日期截斷

        return transformed_data

    def _get_extractor(self, collection):
        EXTRACTOR_MAP = {
            "TEJ_finance_statement": AsyncTEJFinanceStatementDBExtractor,
            "TEJ_self_settlement": AsyncTEJSelfSettlementDBExtractor,
        }

        extractor = EXTRACTOR_MAP.get(collection, None)

        if (extractor is None):
            raise ValueError(
                f"{collection} not a regular argument for collection, only {list(EXTRACTOR_MAP.keys())} available"
            )

        return extractor



from .base import BaseMonthRevenueTransformer
from datetime import datetime
from neurostats_API.utils import StatsProcessor, YoY_Calculator
import pandas as pd

class TWSEMonthlyRevenueTransformer(BaseMonthRevenueTransformer):
    def __init__(self, ticker, company_name, zone):
        super().__init__(ticker, company_name, zone)

        self.data_df = None
    def process_transform(self, fetched_data):
        if (not fetched_data):
            return self._get_empty_structure()
        self.data_df = self._process_data(fetched_data)
        target_month = fetched_data[0]['month']

        self._apply_process_unit_pipeline(
            self.data_df,
            postfix_list = [
                'YoY_1',
                'YoY_3',
                "YoY_5",
                "YoY_10",
                "ratio",
                "percentage",
                "increment",
                "revenue"
            ]
        )

        target_month_df = self.data_df[self.data_df['month'] == target_month]
        annual_month_df = self.data_df[self.data_df['month'] == 12]

        month_revenue_df = self.data_df.pivot(
            index='month', columns='year', values='revenue'
        )
        month_revenue_df = month_revenue_df.sort_index(ascending=False)

        grand_total_df = target_month_df.pivot(
            index='month', columns='year', values='grand_total'
        )
        grand_total_df.rename(
            index={target_month: f"grand_total"}, inplace=True
        )
        month_revenue_df = pd.concat([grand_total_df, month_revenue_df], axis=0)

        annual_total_df = annual_month_df.pivot(
            index='month', columns='year', values='grand_total'
        )

        target_month_df = target_month_df.set_index("year").T

        return_dict = {
            "month_revenue": month_revenue_df[sorted(month_revenue_df.columns, reverse=True)],
            "this_month_revenue_over_years": target_month_df.loc[[
                "revenue", "revenue_increment_ratio", "YoY_1", "YoY_3", "YoY_5", "YoY_10"
            ]],
            "grand_total_over_years": target_month_df.loc[[
                "grand_total", "grand_total_increment_ratio", "grand_total_YoY_1", "grand_total_YoY_3", "grand_total_YoY_5", "grand_total_YoY_10"
            ]],
            "recent_month_revenue": self._get_recent_growth(
                fetched_data, 
                grand_total_dict=annual_total_df.to_dict(), 
                interval=12
            )
        }

        return return_dict

    def _get_recent_growth(self, monthly_data, grand_total_dict, interval=12):
        last_month_data = monthly_data[-(interval + 1): ]  + [{}] * max(0, interval - len(monthly_data) + 1)

        MoMs = [
            YoY_Calculator.cal_growth(this.get('revenue'), last.get('revenue'), delta = 1)
            for last, this in zip(last_month_data[:interval], last_month_data[1 :interval + 1])
        ]

        def safe_accum_yoy(data):
            try:
                year = data['year'] - 1
                total = grand_total_dict[year][12]
                grand_total = data.get('grand_total')
                return f"{round(((grand_total - total) / total) * 100, 2)}%"
            except Exception:
                return None

        recent_month_data = {
            "date": [f"{d.get('year', 0)}/{d.get('month', 0)}" for d in last_month_data[:interval]],
            "revenue": [d.get('revenue') for d in last_month_data[:interval]],
            "MoM": [f"{(m * 100):.2f}%" if isinstance(m, float) else None for m in MoMs],
            "YoY": [d.get('revenue_increment_ratio') for d in last_month_data[:interval]],
            "total_YoY": [d.get('grand_total_increment_ratio') for d in last_month_data[:interval]],
            # accum_YoY
            # accum_YoY 為 Davis提出的定義
            # 2024/6的累計YoY(accum_YoY) 為 2024累計到6月為止的總營收/2023年度總營收
            "accum_YoY": [safe_accum_yoy(d) for d in last_month_data[:interval]]
        }

        df = pd.DataFrame(recent_month_data)
        return df[df['date'] != "0/0"].set_index('date').T


    def _get_empty_structure(self):
        """
        Exception 發生時回傳
        """
        target_date = datetime.today()
        target_year, target_month = target_date.year, target_date.month
        recent_date = [f"{target_year}/{target_month}"]
        for _ in range(11):
            target_year, target_month = (target_year - 1, 12) if target_month == 1 else (target_year, target_month - 1)
            recent_date.append(f"{target_year}/{target_month}")

        def empty_df(index, columns):
            return pd.DataFrame(index=index, columns=columns)

        return {
            "month_revenue": empty_df(
                index=pd.Index(['grand_total', 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1], dtype='object', name='month'),
                columns=pd.Index([f"{target_year - i}" for i in range(10)], dtype=object, name='year')
            ),
            "this_month_revenue_over_years": empty_df(
                index=pd.Index(['revenue', 'revenue_increment_ratio', 'YoY_1', 'YoY_3', 'YoY_5', 'YoY_10'], dtype='object'),
                columns=pd.Index([f"{target_year - i}" for i in range(10)], dtype='int64', name='year')
            ),
            "grand_total_over_years": empty_df(
                index=pd.Index(['grand_total', 'grand_total_increment_ratio', 'grand_total_YoY_1', 'grand_total_YoY_3', 'grand_total_YoY_5', 'grand_total_YoY_10'], dtype='object'),
                columns=pd.Index([f"{target_year - i}" for i in range(10)], dtype='int64', name='year')
            ),
            "recent_month_revenue": empty_df(
                index=pd.Index(['revenue', 'MoM', 'YoY', 'total_YoY', 'accum_YoY'], dtype='object'),
                columns=pd.Index([], dtype = 'object', name = 'date')
            )
        }


from ..base import BaseTransformer
from neurostats_API.utils import StatsProcessor, YoY_Calculator
import pandas as pd


class BaseTEJTransformer(BaseTransformer):

    def __init__(self, ticker, company_name, zone):
        super().__init__(ticker, company_name, zone)
        self.thousand_index_list = [] # 在子類別定義
        self.percent_index_list = [] # 在子類別定義
        self.skip_index = [] # 在子類別定義

    def _process_data_to_tej_format(self, fetched_data, target_key):
        table_dict = {}
        for data in fetched_data:
            year, season, target_dict = data['year'], data['season'], data[
                target_key]
            time_index = f"{year}Q{season}"
            table_dict[time_index] = target_dict

        return table_dict

    def _transform_value(self, data_dict):
        """
        處理千元, %等單位
        """

        data_df = pd.DataFrame.from_dict(data_dict)
        for category, postfix in [(self.thousand_index_list, "千元"),
                                  (self.percent_index_list, "%")]:
            process_list = list(set(data_df.index) & set(category))

            if postfix == "%":
                data_df = data_df.T
                data_df[process_list] = data_df[process_list].map(
                    lambda x: f"{x}%"
                )    # if (not np.isnan(x)) else None)
                data_df = data_df.T
            else:
                data_df.loc[process_list] = data_df.loc[process_list].map(
                    lambda x: StatsProcessor.
                    cal_non_percentage(x, postfix=postfix)
                )
        return data_df.to_dict()

    def _calculate_growth(self, this_value, last_value, delta):
        try:
            return YoY_Calculator.cal_growth(
                this_value, last_value, delta
            ) * 100
        except Exception:
            return None

    def _get_QoQ_data(self, data, use_cal=False):

        if (use_cal):
            new_dict = {}
            for time_index, data_dict in data.items():
                year, season = time_index.split('Q')
                year, season = int(year), int(season)
                last_year, last_season = (year - 1, 4) if season == 1 else (
                    year, season - 1
                )

                pre_data_dict = data.get(f"{last_year}Q{last_season}", {})

                new_dict[time_index] = self._calculate_QoQ_dict(
                    data_dict, pre_data_dict, delta=1
                )
            data = pd.DataFrame.from_dict(new_dict)
            return data

        else:
            data = pd.DataFrame.from_dict(data)
            return data

    def _get_YoY_data(self, data, target_season, use_cal=False):
        if (use_cal):
            new_dict = {}
            year_shifts = [1, 3, 5, 10]

            for time_index, data_dict in data.items():
                year, season = time_index.split('Q')
                year, season = int(year), int(season)

                target_items = []
                for shift in year_shifts:
                    last_year, last_season = (year - shift, season)

                    pre_data_dict = data.get(f"{last_year}Q{last_season}", {})

                    target_items.append(pre_data_dict)

                new_dict[time_index] = self._calculate_YoY_dict(
                    data_dict, target_items, year_shifts = year_shifts
                )

            data = pd.DataFrame.from_dict(new_dict)
            target_season_col = data.columns.str.endswith(f"{target_season}")
            data = data.loc[:, target_season_col]
        
        else:
            data = pd.DataFrame.from_dict(data)
            target_season_col = data.columns.str.endswith(f"{target_season}")
            data = data.loc[:, target_season_col]

        return data

    def _calculate_QoQ_dict(self, this_dict, last_dict, delta):
        temp_dict = {}
        for key in list(this_dict.keys()):
            if key in self.skip_index:
                temp_dict[f"{key}_value"] = this_dict[key]
                continue
            this_value = self._process_value(this_dict[key])

            last_value = last_dict.get(key, None)
            last_value = self._process_value(last_value)

            growth = self._calculate_growth(
                this_value, last_value, delta=delta
            ) if last_value is not None else None

            temp_dict[f"{key}_value"] = this_dict[key]
            temp_dict[f"{key}_growth"] = (f"{growth:.2f}%" if growth else None)

        return temp_dict
    
    def _calculate_YoY_dict(self, this_dict, last_dicts, year_shifts):
        temp_dict = {}
        for last_dict, delta in zip(last_dicts, year_shifts):
            for key in list(this_dict.keys()):
                if key in self.skip_index:
                    temp_dict[f"{key}_value"] = this_dict[key]
                    continue
                this_value = self._process_value(this_dict[key])

                last_value = last_dict.get(key, None)
                last_value = self._process_value(last_value)
                
                growth = self._calculate_growth(
                    this_value, last_value, delta=delta
                ) if last_value is not None else None

                temp_dict[f"{key}_value"] = this_dict[key]
                temp_dict[f"{key}_YoY_{delta}"] = (f"{growth:.2f}%" if growth else None)

        return temp_dict



from .base import BaseDBClient
import os
from pymongo import AsyncMongoClient
from dotenv import load_dotenv

class TEJDBClient(BaseDBClient):
    def __init__(self, mongo_uri):
        """初始化時接收 MongoDB 連接 URI"""
        super().__init__(mongo_uri, zone = 'tw')
        self.db = self.client["company_test"]


from .base import AsyncBaseFetcher
from datetime import datetime, timedelta
from ..db_extractors import (
    AsyncTWSEChipDBExtractor,
    AsyncYFDailyTechDBExtractor,
)
from neurostats_API.async_mode.factory import get_extractor
from neurostats_API.utils import NotSupportedError
from neurostats_API.transformers import TWSEChipTransformer


class AsyncTWSEInstitutionFetcher(AsyncBaseFetcher):

    def __init__(self, ticker, client):
        self.ticker = ticker
        self.transformer_map = {"tw": TWSEChipTransformer}
        try:
            self.extractor = get_extractor(
                "TWSE_chip", ticker, client, fetch_type='I'
            )
            self.daily_extractor = get_extractor("YF_tech", ticker, client)

            company_name = self.extractor.get_company_name()
            zone = self.extractor.get_zone()

            transformer = self.get_transformer(zone)
            self.transformer = transformer(ticker, company_name, zone)

        except NotSupportedError as e:
            raise NotSupportedError(
                f"{self.__class__.__name__} only support {list(self.transformer_map.keys())} companies now, {ticker} is not available"
            ) from e

    async def query_data(
        self, start_date=None, end_date=None, get_latest=False
    ):
        fetched_data = await self.extractor.query_data(
            start_date, end_date, get_latest
        )
        tech_data = await self.daily_extractor.query_data(
            start_date, end_date, get_latest
        )

        fetched_data = {"institution_trading": fetched_data}
        transformed_data = self.transformer.process_transform(
            tech_data=tech_data, **fetched_data
        )

        return transformed_data

    async def query_data_annual(self):
        """
        取一年內的資料
        """
        end_date = datetime.today()
        start_date = end_date - timedelta(days=365)

        fetched_data = await self.query_data(
            start_date, end_date, get_latest=False
        )

        return fetched_data
